import asyncio
import unittest
from os import path

import dill

from deltalanguage.data_types import DFloat, DInt, DUInt
from deltalanguage.runtime import serialize_graph
from deltalanguage.wiring import DeltaGraph
from deltasimulator.build_tools import BuildArtifact
from deltasimulator.build_tools.environments import VerilatorEnv

from test._utils import DUT1, print_then_exit, return_1000

class TestVerilator(unittest.TestCase):

    def test_float_width_sum(self):
        with VerilatorEnv() as env:
            self.assertEqual(env.as_c_type(dill.dumps(DFloat())), "sc_bv<32>")

    def test_int_size(self):
        with VerilatorEnv() as env:
            self.assertEqual(env.as_c_type(dill.dumps(DInt())), "sc_bv<32>")

    def test_unit_size(self):
        with VerilatorEnv() as env:
            self.assertEqual(env.as_c_type(dill.dumps(DUInt())), "sc_bv<32>")

    def test_migen_node(self):
        with DeltaGraph() as test_graph:
            c1 = DUT1(tb_num_iter=2000, name='counter1').call(i1=return_1000())
            print_then_exit(c1.o1)
        _, serialised = serialize_graph(test_graph)
        top_v = BuildArtifact(name=f"{serialised.nodes[1].name}",
                              data=serialised.bodies[1].migen.verilog.encode("utf-8"))
        with VerilatorEnv() as env:
            build_artifacts = env.verilate(top_v)
        asyncio.run(self.assert_build_correct(build_artifacts))

    async def assert_build_correct(self, build_artifacts):
        """Build artifacts and check the compiled code is correct."""
        built = {}
        built["cpp"] = await asyncio.wait_for(build_artifacts["cpp"].data,
                                              timeout=None)
        built["h"] = await asyncio.wait_for(build_artifacts["h"].data,
                                            timeout=None)
        built["ALL.a"] = await asyncio.wait_for(build_artifacts["ALL.a"].data,
                                                timeout=None)
        built["verilated.o"] = await asyncio.wait_for(
            build_artifacts["verilated.o"].data, timeout=None)
        built["init1"] = await asyncio.wait_for(build_artifacts["init"][0].data,
                                               timeout=None)
        built["init2"] = await asyncio.wait_for(build_artifacts["init"][1].data,
                                               timeout=None)
        with open(path.join("test", "data", "verilated_cpp.out"), "rb") as cpp_file:
            self.assertEqual(cpp_file.read(), built["cpp"])
        with open(path.join("test", "data", "verilated_h.out"), "rb") as h_file:
            self.assertEqual(h_file.read(), built["h"])
        with open(path.join("test", "data", "verilated_init1.out"), "rb") as init_file:
            self.assertEqual(init_file.read(), built["init1"])
        with open(path.join("test", "data", "verilated_init2.out"), "rb") as init_file:
            self.assertEqual(init_file.read(), built["init2"])

    def tearDown(self):
        DeltaGraph.clean_stack()


if __name__ == "__main__":
    unittest.main()



import glob
import unittest
import subprocess
from os import path, chmod, environ
import shutil
import subprocess
import sys
from tempfile import TemporaryDirectory, NamedTemporaryFile

from deltalanguage.data_types import DInt, DSize, NoMessage, make_forked_return, DOptional, DBool
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaGraph,
                                  template_node_factory,
                                  Interactive,
                                  PyInteractiveNode,
                                  placeholder_node_factory)
from deltalanguage.runtime import DeltaRuntimeExit, deserialize_graph, serialize_graph
from deltalanguage.lib.quantum_simulators import QiskitQuantumSimulator, ProjectqQuantumSimulator
from deltalanguage.lib.hal import command_creator, HardwareAbstractionLayerNode
from deltalanguage.lib import StateSaver
from deltasimulator.lib import build_graph

from test._utils import (DUT1,
                         MigenIncrementer,
                         add,
                         print_then_exit,
                         print_then_exit_64_bit,
                         exit_if_true,
                         return_1000,
                         send_gates_list_then_exit)


ExpT, ExpVal = make_forked_return(
    {'num_out': DInt(DSize(32)), 'val_out': DBool()})


class TestExecution(unittest.TestCase):

    def check_executes_graph(self, graph, expect=None, files=[], reqs=[]):
        _, program = serialize_graph(graph, name="dut", files=files)
        self.check_executes(program, expect=expect, files=files, reqs=reqs)

    def check_executes_file(self, file, expect=None, files=[], reqs=[]):
        with open(file, "rb") as df_file:
            program = deserialize_graph(df_file.read())
        self.check_executes(program, expect=expect, files=files, reqs=reqs)

    def check_executes(self, program, expect: str = None, files=[], reqs=[]):
        """Build SystemC program and executes them in temp directory."""
        self.files = set(
            [file for pattern in files for file in glob.glob(pattern)])
        for file in self.files:
            shutil.move(file, f"{file}_temp")
        shutil.rmtree("__pycache__", ignore_errors=True)
        self.reqs = reqs
        with TemporaryDirectory() as build_dir:
            build_graph(program, main_cpp=path.join(path.dirname(__file__), "main.cpp"),
                        build_dir=build_dir)
            # Setting the permission to run the file
            chmod(f"{build_dir}/main", 0o777)
            try:
                # We disable the SYSTEMC Banner to clear the output by setting the
                # SYSTEMC_DISABLE_COPYRIGHT_MESSAGE
                env = dict(environ, SYSTEMC_DISABLE_COPYRIGHT_MESSAGE="1")
                _proc = subprocess.run("./main",
                                       cwd=build_dir,
                                       shell=True,
                                       check=True,
                                       stdout=subprocess.PIPE,
                                       env=env)
            except subprocess.CalledProcessError as e:
                print(f"Failure in running: {e.returncode} - {e.output}")
                raise e
            if expect:
                self.assertRegex(_proc.stdout.decode(), expect,
                                 f"{_proc.stdout.decode()} does not match with the provided regex")

    def test_add(self):
        with DeltaGraph(name="test_add") as test_graph:
            print_then_exit(n=add(a=2, b=3))

        self.check_executes_graph(test_graph, "5")

    def test_add_64_bit(self):
        @DeltaBlock()
        def return_1() -> DInt(DSize(64)):
            return 1

        @DeltaBlock()
        def return_2() -> DInt(DSize(64)):
            return 2

        @DeltaBlock(allow_const=False)
        def add_64_bit(a: DInt(DSize(64)), b: DInt(DSize(64))) -> DInt(DSize(64)):
            return a+b

        with DeltaGraph(name="test_add_64_bit") as test_graph:
            print_then_exit_64_bit(n=add_64_bit(a=return_1(), b=return_2()))

        self.check_executes_graph(test_graph, "3")

    def test_and(self):
        @DeltaBlock(allow_const=False)
        def bool_and(a: bool, b: bool) -> bool:
            return a and b

        @DeltaBlock(allow_const=False)
        def print_then_exit_bool(x: bool) -> NoMessage:
            print(x)
            raise DeltaRuntimeExit

        with DeltaGraph(name="test_and") as test_graph:
            print_then_exit_bool(x=bool_and(a=True, b=False))

        self.check_executes_graph(test_graph, "False")

    def test_forked(self):
        ForkedReturnT, ForkedReturn = make_forked_return({'a': int, 'b': int})

        @DeltaBlock(allow_const=False)
        def add_2_add_3(n: int) -> ForkedReturnT:
            return ForkedReturn(a=n+2, b=n+3)

        with DeltaGraph(name="test_forked") as test_graph:
            ab = add_2_add_3(n=1)
            print_then_exit(n=add(a=ab.a, b=ab.b))

        self.check_executes_graph(test_graph, "7")

    def test_interactive_one_in_one_out(self):
        @Interactive(in_params={"num": int}, out_type=int, name="interactive_one_in_one_out")
        def interactive_func(node: PyInteractiveNode):
            for _ in range(2):
                num = node.receive("num")
                print(f"received num: {num}")
            node.send(num + 1)

        with DeltaGraph(name="test_interactive_one_in_one_out") as test_graph:
            print_then_exit(n=interactive_func.call(num=add(a=2, b=3)))

        self.check_executes_graph(
            test_graph, "received num: 5\\nreceived num: 5\\n6")

    def test_interactive_one_in_two_out(self):
        @Interactive(in_params={"num": DInt(DSize(32))}, out_type=ExpT, name="interactive_one_in_two_out")
        def interactive_func(node: PyInteractiveNode):
            for _ in range(2):
                num = node.receive("num")
                print(f"received num: {num}")
            node.send(ExpVal(num_out=None, val_out=False))
            node.send(ExpVal(num_out=14, val_out=False))

        with DeltaGraph(name="interactive_one_in_two_out") as test_graph:
            int_func = interactive_func.call(num=4, opt_val=True)
            exit_if_true(cond=int_func.val_out)
            print_then_exit(n=int_func.num_out)
        self.check_executes_graph(
            test_graph, "received num: 4\\nreceived num: 4\\n14")

    def test_interactive_two_in_one_out(self):
        @Interactive(in_params={"num": DInt(DSize(32)), "opt_val": DOptional(DBool())}, out_type=DBool(), name="interactive_two_in_one_out")
        def interactive_func(node: PyInteractiveNode):
            for _ in range(2):
                num = node.receive("num")
                opt_val = node.receive("opt_val")
                print(f"received opt_val: {opt_val}")
                print(f"received num: {num}")
            node.send(True)

        with DeltaGraph(name="interactive_two_in_one_out") as test_graph:
            int_func = interactive_func.call(num=4, opt_val=True)
            exit_if_true(cond=int_func)
        self.check_executes_graph(
            test_graph, "received opt_val: True\\nreceived num: 4\\nreceived opt_val: True\\nreceived num: 4")

    def test_interactive_two_in_two_out(self):
        @Interactive(in_params={"num": DInt(DSize(32)), "opt_val": DOptional(DBool())}, out_type=ExpT, name="interactive_two_in_two_out")
        def interactive_func(node: PyInteractiveNode):
            for _ in range(2):
                num = node.receive("num")
                opt_val = node.receive("opt_val")
                print(f"received opt_val: {opt_val}")
                print(f"received num: {num}")
            node.send(ExpVal(num_out=None, val_out=False))
            node.send(ExpVal(num_out=14, val_out=False))

        with DeltaGraph(name="interactive_two_in_two_out") as test_graph:
            int_func = interactive_func.call(num=4, opt_val=True)
            exit_if_true(cond=int_func.val_out)
            print_then_exit(n=int_func.num_out)
        self.check_executes_graph(
            test_graph, "received opt_val: True\\nreceived num: 4\\nreceived opt_val: True\\nreceived num: 4\\n14")

    def test_splitter(self):
        with DeltaGraph(name="test_splitter") as test_graph:
            n = add(a=2, b=3)
            print_then_exit(n=add(a=n, b=n))

        self.check_executes_graph(test_graph, "10")

    def test_migen(self):
        with DeltaGraph("test_migen_wiring") as test_graph:
            c1 = DUT1(tb_num_iter=2000, name='counter1').call(i1=return_1000())
            c2 = DUT1(tb_num_iter=2000, name='counter2').call(i1=c1.o1)
            print_then_exit(c2.o1)

        self.check_executes_graph(test_graph, "10")

    def test_migen_template(self):
        with DeltaGraph("test_migen_template") as test_graph:
            c1 = DUT1(tb_num_iter=2000, name='counter1').call(i1=return_1000())
            c2 = DUT1(tb_num_iter=2000, name='counter2').call(
                i1=template_node_factory(return_type=int, a=c1.o1))
            print_then_exit(c2.o1)

        self.check_executes_graph(test_graph, "10")

    def test_migen_python(self):
        @DeltaBlock(allow_const=False)
        def exit_if_6_else_inc(n: int) -> int:
            print(n)
            if n == 6:
                raise DeltaRuntimeExit
            else:
                return n+1

        with DeltaGraph("test_migen_python") as test_graph:
            ph = placeholder_node_factory()
            c1 = MigenIncrementer(
                tb_num_iter=2000, name='counter1', vcd_name="/workdir/counter1.vcd").call(i1=ph)
            ex = exit_if_6_else_inc(c1.o1)
            ph.specify_by_node(ex)
        self.check_executes_graph(test_graph, "0\\n2\\n4\\n6\\n")

    def test_loop_with_ProjectQ(self):
        with DeltaGraph("test_loop_with_ProjectQ") as test_graph:
            # set up placeholders
            ph_hal_result = placeholder_node_factory()

            int_func = send_gates_list_then_exit.call(
                measurement=ph_hal_result)

            projectQ = HardwareAbstractionLayerNode(ProjectqQuantumSimulator(
                register_size=2)).accept_command(command=int_func)
            # tie up placeholders
            ph_hal_result.specify_by_node(projectQ)

        self.check_executes_graph(test_graph, "Measurement: 429490176[0-1]")

    @unittest.skip("Occasionally fails due to a segfault.")
    def test_loop_with_Qiskit(self):
        with DeltaGraph("test_loop_with_Qiskit") as test_graph:
            # set up placeholders
            ph_hal_result = placeholder_node_factory()

            int_func = send_gates_list_then_exit.call(
                measurement=ph_hal_result)

            qiskit = HardwareAbstractionLayerNode(QiskitQuantumSimulator(
                register_size=2, seed=2)).accept_command(command=int_func)
            # tie up placeholders
            ph_hal_result.specify_by_node(qiskit)

        self.check_executes_graph(test_graph)

    def test_state_saver(self):
        store = StateSaver(int)
        with DeltaGraph("test_state_saver") as test_graph:
            c1 = DUT1(tb_num_iter=2000, name='counter1').call(i1=return_1000())
            store.save_and_exit(c1.o1)
        self.check_executes_graph(test_graph)

    def test_state_saver_save_to_file(self):
        with NamedTemporaryFile(mode="w+") as f:
            store = StateSaver(int, filename=f.name)
            with DeltaGraph("test_state_saver_save_to_file") as test_graph:
                store.save_and_exit(return_1000())
            self.check_executes_graph(test_graph)
            f.seek(0)
            contents = f.read()
            expected = "1000"
            self.assertEqual(contents, f"{expected}\n")

    def test_read_file(self):
        filename = path.join("test", "data", "test_add_64_bit_cpp.out")

        @DeltaBlock(allow_const=False)
        def read_txt() -> NoMessage:
            with open(filename, "r") as txt_file:
                print(txt_file.read())
            raise DeltaRuntimeExit
        with DeltaGraph("test_read_file") as test_graph:
            read_txt()
        self.check_executes_graph(test_graph, files=[filename])

    def test_read_multiple_files(self):
        filenames = [path.join("test", "data", "test_add_64_bit_cpp.out"),
                     path.join("test", "data", "test_add_64_bit_h.out")]

        @DeltaBlock(allow_const=False)
        def read_txt() -> NoMessage:
            for filename in filenames:
                with open(filename, "r") as txt_file:
                    print(txt_file.read())
            raise DeltaRuntimeExit
        with DeltaGraph("test_read_multiple_files") as test_graph:
            read_txt()
        self.check_executes_graph(test_graph, files=filenames)

    def test_read_pattern_files(self):
        pattern = path.join("test", "data", "test_add_*.out")

        @DeltaBlock(allow_const=False)
        def read_txt() -> NoMessage:
            for filename in glob.glob(pattern):
                with open(filename, "r") as txt_file:
                    print(txt_file.read())
            raise DeltaRuntimeExit
        with DeltaGraph("test_read_pattern_files") as test_graph:
            read_txt()
        self.check_executes_graph(test_graph, files=[pattern])

    def test_python_file(self):
        import test.data.code_for_test as code_for_test

        @DeltaBlock(allow_const=False)
        def run_code() -> NoMessage:
            print(code_for_test.x)
            raise DeltaRuntimeExit
        with DeltaGraph("test_python_file") as test_graph:
            run_code()
        files = [path.join("test", "data", "code_for_test.py"),
                 path.join("test", "__init__.py"),
                 path.join("test", "_utils.py")]
        self.check_executes_graph(test_graph, expect="5", files=files)

    def test_package_dependency(self):
        """If a program requires a package that is not installed, use pip
        to import the correct package.
        """
        df_filename = path.join("test", "data", "permutation.df")
        # If we need to rebuild the .df file for some reason, run
        # pip install permutation
        # and run this test with build_df_file set to True.
        # This will create a new .df file. Then run
        # pip uninstall -y permutation
        # and set build_df_file to False to run the actual test.
        build_df_file = False
        if build_df_file:
            from permutation import Permutation

            @DeltaBlock(allow_const=False)
            def print_cycles() -> NoMessage:
                print(Permutation(2, 1, 4, 5, 3).to_cycles())
                raise DeltaRuntimeExit
            with DeltaGraph("test_package_dependency") as test_graph:
                print_cycles()
            serialized, _ = serialize_graph(
                test_graph, name="dut", requirements=["permutation"])
            with open(df_filename, "wb") as df_file:
                df_file.write(serialized)
        else:
            self.check_executes_file(
                df_filename, expect="[(1, 2), (3, 4, 5)]", reqs=["permutation"])

    def test_incompatible_dependencies(self):
        """If a program requires incompatible dependencies, throw an error
        """
        df_filename = path.join("test", "data", "incompatible.df")
        # If we need to rebuild the .df file for some reason, run
        # pip install permutation
        # and run this test with build_df_file set to True.
        # This will create a new .df file. Then run
        # pip uninstall -y permutation
        # and set build_df_file to False to run the actual test.
        build_df_file = False
        if build_df_file:
            from permutation import Permutation

            @DeltaBlock(allow_const=False)
            def print_cycles() -> NoMessage:
                print(Permutation(2, 1, 4, 5, 3).to_cycles())
                raise DeltaRuntimeExit
            with DeltaGraph("test_incompatible_dependencies") as test_graph:
                print_cycles()
            serialized, _ = serialize_graph(test_graph,
                                            name="dut",
                                            requirements=["permutation<=0.2.0",
                                                          "permutation==0.3.0"])
            with open(df_filename, "wb") as df_file:
                df_file.write(serialized)
        else:
            with self.assertRaises(RuntimeError):
                self.check_executes_file(
                    df_filename, expect="[(1, 2), (3, 4, 5)]")

    @unittest.skip("This test currently fails in a new container.")
    def test_constant_dependency(self):
        """Example where a constant node has a dependency.

        .. todo::
            Currently this test fails if run in a completely new container.
            This is because pip cannot install to a running program, and
            constant nodes are currently deserialised during the build
            process. This will be fixed in future work.
        """
        df_filename = path.join("test", "data", "constant.df")
        # If we need to rebuild the .df file for some reason, run
        # pip install permutation
        # and run this test with build_df_file set to True.
        # This will create a new .df file. Then run
        # pip uninstall -y permutation
        # and set build_df_file to False to run the actual test.
        build_df_file = False
        if build_df_file:
            from permutation import Permutation

            @DeltaBlock(allow_const=False)
            def num_cycles() -> DInt():
                return len(Permutation(2, 1, 4, 5, 3).to_cycles())
            with DeltaGraph("test_constant_dependency") as test_graph:
                print_then_exit(num_cycles())
            serialized, _ = serialize_graph(test_graph,
                                            name="dut",
                                            requirements=["permutation"])
            with open(df_filename, "wb") as df_file:
                df_file.write(serialized)
        else:
            self.check_executes_file(
                df_filename, expect="2", reqs=["permutation"])

    def setUp(self):
        self.files = []
        self.reqs = []

    def tearDown(self):
        DeltaGraph.clean_stack()
        for file in self.files:
            shutil.move(f"{file}_temp", file)
        for req in self.reqs:
            subprocess.check_call(
                [sys.executable, '-m', 'pip', 'uninstall', '-y', req])


if __name__ == "__main__":
    unittest.main()



import asyncio
import unittest
from os import path

from deltalanguage.data_types import DInt, DSize, NoMessage, make_forked_return
from deltalanguage.wiring import DeltaBlock, DeltaGraph, template_node_factory, Interactive, PyInteractiveNode
from deltalanguage.runtime import DeltaRuntimeExit, serialize_graph
from deltasimulator.lib import generate_wiring, _wait_for_build

from test._utils import (DUT1,
                         add,
                         add_const,
                         print_then_exit,
                         print_then_exit_64_bit,
                         return_1000)


class TestWiring(unittest.TestCase):

    def check_build(self, test_graph):
        """Build SystemC program and run tests."""
        _, program = serialize_graph(test_graph)
        _, _, wiring = generate_wiring(program)
        asyncio.run(self.assert_build_correct(wiring, test_graph.name))

    async def assert_build_correct(self, wiring, prog_name):
        """Build artifacts and check the compiled code is correct.
        Artifacts related to sc_main will not exist if there are
        template nodes.
        """
        built = {}
        if f"{prog_name}.cpp" in wiring:
            built[f"{prog_name}.cpp"] = await asyncio.wait_for(wiring[f"{prog_name}.cpp"].data, timeout=None)
            with open(path.join("test", "data", f"{prog_name}_cpp.out"),
                      "rb") as cpp_file:
                ref = cpp_file.read().decode("utf-8")
                gen = built[f"{prog_name}.cpp"].decode("utf-8")
                self.assertMultiLineEqual(gen, ref)
        built[f"{prog_name}.a"] = await asyncio.wait_for(wiring[f"{prog_name}.a"].data,
                                                         timeout=None)
        built[f"{prog_name}.h"] = await asyncio.wait_for(wiring[f"{prog_name}.h"].data, timeout=None)
        mdiff = self.maxDiff
        self.maxDiff = None
        with open(path.join("test", "data", f"{prog_name}_h.out"), "rb") as h_file:
            ref = h_file.read().decode("utf-8")
            gen = built[f"{prog_name}.h"].decode("utf-8")
            self.assertMultiLineEqual(gen, ref)
        if f"{prog_name}.o" in wiring:
            built[f"{prog_name}.o"] = await asyncio.wait_for(wiring[f"{prog_name}.o"].data, timeout=None)
        if f"{prog_name}" in wiring:
            built[f"{prog_name}"] = await asyncio.wait_for(wiring[f"{prog_name}"].data, timeout=None)
        self.maxDiff = mdiff
        self.maxDiff = mdiff

    def test_add(self):
        with DeltaGraph(name="test_add") as test_graph:
            print_then_exit(n=add(a=2, b=3))

        self.check_build(test_graph)

    def test_add_64_bit(self):
        @DeltaBlock()
        def return_1() -> DInt(DSize(64)):
            return 1

        @DeltaBlock()
        def return_2() -> DInt(DSize(64)):
            return 2

        @DeltaBlock(allow_const=False)
        def add_64_bit(a: DInt(DSize(64)), b: DInt(DSize(64))) -> DInt(DSize(64)):
            return a+b

        with DeltaGraph(name="test_add_64_bit") as test_graph:
            print_then_exit_64_bit(n=add_64_bit(a=return_1(), b=return_2()))

        self.check_build(test_graph)

    def test_and(self):
        @DeltaBlock(allow_const=False)
        def bool_and(a: bool, b: bool) -> bool:
            return a and b

        @DeltaBlock(allow_const=False)
        def print_then_exit_bool(x: bool) -> NoMessage:
            print(x)
            raise DeltaRuntimeExit

        with DeltaGraph(name="test_and") as test_graph:
            print_then_exit_bool(x=bool_and(a=True, b=False))

        self.check_build(test_graph)

    def test_forked(self):
        ForkedReturnT, ForkedReturn = make_forked_return({'a': int, 'b': int})

        @DeltaBlock(allow_const=False)
        def add_2_add_3(n: int) -> ForkedReturnT:
            return ForkedReturn(a=n+2, b=n+3)

        with DeltaGraph(name="test_forked") as test_graph:
            ab = add_2_add_3(n=1)
            print_then_exit(n=add(a=ab.a, b=ab.b))

        self.check_build(test_graph)

    def test_interactive(self):
        @Interactive(in_params={"num": int}, out_type=int, name="interactive")
        def interactive_func(node: PyInteractiveNode):
            for _ in range(10):
                num = node.receive()["num"]
                print(f"received num: {num}")
            node.send(num + 1)

        with DeltaGraph(name="test_interactive") as test_graph:
            print_then_exit(n=interactive_func.call(num=add(a=2, b=3)))

        self.check_build(test_graph)


    def test_splitter(self):
        with DeltaGraph(name="test_splitter") as test_graph:
            n = add(a=2, b=3)
            print_then_exit(n=add(a=n, b=n))

        self.check_build(test_graph)

    def test_migen(self):
        with DeltaGraph("test_migen_wiring") as test_graph:
            c1 = DUT1(tb_num_iter=2000, name='counter1').call(i1=return_1000())
            c2 = DUT1(tb_num_iter=2000, name='counter2').call(i1=c1.o1)
            print_then_exit(c2.o1)

        self.check_build(test_graph)

    def test_python_template(self):
        with DeltaGraph("test_python_template") as test_graph:
            print_then_exit(template_node_factory(return_type=int, a=1, b=2))

        self.check_build(test_graph)

    def test_migen_template(self):
        with DeltaGraph("test_migen_template") as test_graph:
            c1 = DUT1(tb_num_iter=2000, name='counter1').call(i1=return_1000())
            c2 = DUT1(tb_num_iter=2000, name='counter2').call(
                i1=template_node_factory(return_type=int, a=c1.o1))
            print_then_exit(c2.o1)

        self.check_build(test_graph)

    def test_const_graph(self):
        with DeltaGraph("test_const_graph") as test_graph:
            add_const(a=5, b=6)

        with self.assertRaises(RuntimeError):
            self.check_build(test_graph)

    def tearDown(self):
        DeltaGraph.clean_stack()


if __name__ == "__main__":
    unittest.main()



"""Summary information about the package
    Attributes
    ----------
    __license__ : str
        Short description of our license.
    __copyright__ : str
        Copyright notice
    __url__ : str
        Project URL
    __contributors__ : str
        Single-line list of all project contributors.
    __contributors_lines__ : str
        Multi-line list of all project contributors.
    __email__ : str
        Contact email for the project.
    __version__ : str
        Package version in format ``<major>.<minor>.<micro>``
    __version_description__ : str
        Short description for the current version.
    __short_description__ : str
        Short summary of the package.
    __doc__ : Str
        Longer description of the package.
    __platforms__ : List[str]
        List of supported platforms
"""

__all__ = [
    "__license__",
    "__copyright__",
    "__url__",
    "__download_url__",
    "__docs_url__",
    "__contributors__",
    "__contributors_lines__",
    "__email__",
    "__version__",
    "__version_description__",
    "__short_description__",
    "__doc__",
    "__platforms__"
]

__license__ = "MIT License with Commons Clause"
__copyright__ = "Copyright (C) 2020 River Lane Research Ltd"


# Source URL
__url__ = "https://github.com/riverlane/deltasimulator"
# Package Hosting URL
__download_url__ = "https://pypi.org/project/deltasimulator/#files"
# Docs Hosting Url
__docs_url__ = "https://riverlane.github.io/deltasimulator"

contributors = ["Kenton Barnes",
                "Anton Buyskikh",
                "Marco Ghibaudi",
                "Alex Moylett",
                "Tom Parks",
                "Jan Snoeijs"]
__contributors__ = ", ".join(contributors)
__contributors_lines__ = "\n".join(contributors)
__email__ = "deltaflow@riverlane.com"

version_info = (0, 4, 1)
"""Tuple[int, int, int] : version information
The three components of the version:
``major``, ``minor`` and ``micro``: Module level variable documented inline.
"""
__version__ = ".".join(map(str, version_info[:3]))
__version_description__ = "Deltasimulator MVP Public Release."

__short_description__ = "Runs Deltaflow programs on virtual platforms"
__doc__ = """
    Deltasimulator is a build system developed by Riverlane. The system can be
    used to compile SystemC modules for nodes in a Deltaflow graph and wire the
    nodes up. From there a complete runtime executable can be compiled, or the
    graph can be connected to a larger platform representing physical devices
    such as the ARTIQ platform via a Hardware Abstraction Layer (HAL).
    """

__platforms__ = ['Ubuntu 20.04']



import asyncio
import traceback
import logging
log = logging.getLogger(__name__)

async def wait(flag):
    """Often, we want to sequence multiple actions that need to happen
    after a single coro.

    Wait allows you to purely block on a coro completing whilst discarding the
    result, and allows this to happen multiple times.

    Parameters
    ----------

    flag : Coroutine
        The flag that we are waiting to terminate.    
    """
    fut = asyncio.ensure_future(flag)
    try:
        await fut
    except RuntimeError:
        log.debug(traceback.format_exc())
        pass
    return

async def multiple_waits(flags):
    """Same as :meth:`wait`, but for waiting on multiple coroutines.

    Parameters
    ----------

    flags : list
        A list of :class:`Coroutine` objects to wait for.
    """
    for flag in flags:
        await wait(flag)
    return



from .fileio import *
from .cogify import cogify
from .environment import *
from .utils import *

# set up build system logging.
import logging
build_log = logging.getLogger(__name__)
console = logging.StreamHandler()
build_log.addHandler(console)
build_log.setLevel(logging.WARNING)



"""A build module is a managed enviroment for safe and composable use
of compilation tools.

A specific module will provide a Enviroment. By using an enviroment as a context,
you get access to the tools in that enviromet and can create chains of
async transformations to impliment a compilation process.
"""

from .verilator import VerilatorEnv
from .pythonator import PythonatorEnv
from .wiringenv import WiringEnv
from .cppenv import CPPEnv
from .host import HostEnv



import os.path as path

import dill

from deltalanguage.wiring import PyConstBody

from .cppenv import CPPEnv
from .pythonator import PythonatorEnv
from .verilator import VerilatorEnv
from deltasimulator.build_tools import multiple_waits
from deltasimulator.build_tools.cogify import cogify
from deltasimulator.build_tools.fileio import BuildArtifact, write_futures


class WiringEnv(CPPEnv):
    """Environment that wires the completed graph together.

    When run, a .a archive is produced containing all built node objects,
    as well as a header file containing all node headers
    and a SystemC module for the wiring of nodes.

    If there are no template nodes in the graph then a main runtime is
    also produced and built into an executable. This can be run with
    either no arguments in which case the program runs until
    a :exc:`DeltaRuntimeExit` occurs, or with an optional integer argument
    in which case the program runs for the given number of nanoseconds.

    Parameters
    ----------
    capnp_nodes : list
        Nodes in the graph, described as capnp objects
    capnp_bodies : list
        Bodies for the nodes, described as capnp objects
    node_headers : list
        Header files for the nodes, wrapped in :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects.
    node_objects : list
        Built binary objects for the nodes.
        For Python nodes these are .o files, and for Migen these are .a
        archives. In both cases the files are wrapped in
        :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects.
    verilated_o : Optional[BuildArtifact]
        The content of the verilated.o file, which is required to link objects
        built using Verilator. Note that this is the same regardless of the
        node it was built with, so only a single copy is required.
    prog_name : Optional[str]
        The name of this program, by default "main"


    .. note::
        Connections are note made between pairs of constant Python nodes.
        This is because these nodes are evaluated during the build process
        and therefore require no input during runtime.

    .. note::
        For wiring up template nodes, the wires connecting to template
        nodes are exposed publicly so that these can be connected to their
        actual implementations. Note that connections between template nodes
        are not wired up, as we do not know the communication protocol
        between a pair of template nodes.

    .. note::
        Python nodes and Migen nodes have different communication protocols.
        The header file includes SystemC modules
        :cpp:class:`MigenToPython` and :cpp:class:`PythonToMigen` which
        convert between these different protocols. These can also be used
        to convert the signals between a node in the graph and the actual
        implementation of a template node if needed.
    """

    def __init__(self, capnp_nodes, capnp_bodies, node_headers, node_objects, verilated_o=None, prog_name="main"):
        super().__init__()
        if prog_name:
            self._prog_name = prog_name
        else:
            self._prog_name = "main"
        self._capnp_nodes = capnp_nodes
        self._capnp_bodies = capnp_bodies
        self._node_headers = node_headers
        self._node_objects = node_objects
        if verilated_o:
            self._node_objects.append(verilated_o)
        self._get_node_types()

    @staticmethod
    def get_wire_name(wire):
        """Gets the name of a wire.

        Parameters
        ----------
        wire
            capnp object describing a wire in the graph.

        Returns
        -------
        str
            The name of the wire used by this environment.
            The format used is `{src_node}_{src_port}_{dest_node}_{dest_port}`,
            where `src_node` & `dest_node` are the indexes of the source
            and destination node in the list of nodes, and `src_port`
            & `dest_port` are the indexes of the ports in their respective
            nodes list of ports.
        """
        return "_".join(["wire", str(wire.srcNode), str(wire.srcOutPort), str(wire.destNode), str(wire.destInPort)])

    def get_template_wire_name(self, wire, body="python", direction="out"):
        """Gets standardised name of a wire connecting to a template node.

        These names are a different formats as they are public, so the name
        is more human-readable.

        Parameters
        ----------
        wire
            capnp object describing a wire in the graph.
        body : Optional[str]
            The body-type of the non-template node, by default "python"
        direction : Optional[str]
            The direction of the wire, by default "out"

        Returns
        -------
        str
            The name of the wire used by this environment.
            The format used is `{node_name}_{wire_name}` where
            `node_name` is the name of the node and `wire_name` is the
            name of the SystemC module's port.
            See :meth:`PythonatorEnv.get_sysc_port_name` and
            :meth:`VerilatorEnv.get_sysc_port_name` for the formats of
            `wire_name`.
        """
        if direction == "out":
            node = self._capnp_nodes[wire.srcNode]
            port = node.outPorts[wire.srcOutPort]
        elif direction == "in":
            node = self._capnp_nodes[wire.destNode]
            port = node.inPorts[wire.destInPort]
        if body == "python":
            return f"{node.name}_{PythonatorEnv.get_sysc_port_name(port)}"
        elif body == "interactive":
            return f"{node.name}_{PythonatorEnv.get_sysc_port_name(port)}"
        elif body == "migen":
            return (f"{node.name}_{wire_name}" for wire_name in VerilatorEnv.get_sysc_port_name(port, direction=direction))

    def _get_top_name(self):
        """Gets name of top module.

        Returns
        -------
        str
            The name of the module. The format is a capitalised version
            of the program's name.
        """
        return "_".join([word.capitalize() for word in self._prog_name.split("_")])

    def _get_node_types(self):
        """Determines which nodes are python, migen or template"""
        self._py_nodes = []
        self._migen_nodes = []
        self._has_templates = False
        for node in self._capnp_nodes:
            if node.body == -1:
                self._has_templates = True
            else:
                body_type = self._capnp_bodies[node.body].which()
                if body_type in ["python", "interactive"]:
                    self._py_nodes.append(node)
                elif body_type == "migen":
                    self._migen_nodes.append(node)

    def _get_adaptors(self, capnp_graph):
        """Determines which conversions are needed between nodes"""
        self._py_to_py = []
        self._py_to_migen = []
        self._py_to_template = []
        self._migen_to_migen = []
        self._migen_to_py = []
        self._migen_to_template = []
        self._template_to_py = []
        self._template_to_migen = []
        num_const_wires = 0
        for wire in capnp_graph:
            src_body_index = self._capnp_nodes[wire.srcNode].body
            if (src_body_index == -1):
                src_body_type = "template"
            else:
                src_body_type = self._capnp_bodies[src_body_index].which()
            dest_body_index = self._capnp_nodes[wire.destNode].body
            if (dest_body_index == -1):
                dest_body_type = "template"
            else:
                dest_body_type = self._capnp_bodies[dest_body_index].which()
            if (src_body_type in ["python", "interactive"]) and (dest_body_type in ["python", "interactive"]):
                if src_body_type == "interactive":
                    src_body = type(dill.loads(self._capnp_bodies[src_body_index].interactive.dillImpl))
                else:
                    src_body = type(dill.loads(self._capnp_bodies[src_body_index].python.dillImpl))
                if dest_body_type == "interactive":
                    dest_body = type(dill.loads(self._capnp_bodies[dest_body_index].interactive.dillImpl))
                else:
                    dest_body = type(dill.loads(self._capnp_bodies[dest_body_index].python.dillImpl))
                if (src_body is PyConstBody) and (dest_body is PyConstBody):
                    # Don't wire constant-to-constant nodes
                    num_const_wires += 1
                else:
                    self._py_to_py.append(wire)
            elif (src_body_type in ["python", "interactive"]) and (dest_body_type == "migen"):
                self._py_to_migen.append(wire)
            elif (src_body_type in ["python", "interactive"]) and (dest_body_type == "template"):
                self._py_to_template.append(wire)
            elif (src_body_type == "migen") and (dest_body_type == "migen"):
                self._migen_to_migen.append(wire)
            elif (src_body_type == "migen") and (dest_body_type in ["python", "interactive"]):
                self._migen_to_py.append(wire)
            elif (src_body_type == "migen") and (dest_body_type == "template"):
                self._migen_to_template.append(wire)
            elif (src_body_type == "template") and (dest_body_type in ["python", "interactive"]):
                self._template_to_py.append(wire)
            elif (src_body_type == "template") and (dest_body_type == "migen"):
                self._template_to_migen.append(wire)
        if num_const_wires == len(capnp_graph) and num_const_wires > 0:
            raise RuntimeError("Graph cannot consist of only constant nodes.")

    async def _make_main(self):
        """Makes the C++ file containing :cpp:func:`sc_main`.

        Returns
        -------
        bool
            Returns True when complete.


        .. warning::
            Only run this method if there are no template nodes in the graph.
            Otherwise the graph will be incomplete and fail to run correctly.
        """
        main_tmpl = """
        #include <systemc>
        #include <iostream>
        #include <string>
        /*[[[cog
            cog.outl(f'#include "{self._prog_name}.h"')
        ]]]*/
        //[[[end]]]
        #include "Python.h"
        using namespace sc_core;
        using namespace std;
        int sc_main(int argc, char* argv[]) {
            Py_UnbufferedStdioFlag = 1;
            Py_Initialize();
            // Adds required components: trace, clock, reset
            /*[[[cog
                cog.outl(f'sc_trace_file *Tf = sc_create_vcd_trace_file("{self._prog_name}");')
                cog.outl(f'{self._get_top_name()} {self._prog_name}("{self._prog_name}", Tf);')
                cog.outl(f'sc_clock clk("clk", sc_time(1, SC_NS)); sc_trace(Tf, clk, "clk");')
                cog.outl(f'sc_signal<bool> rst; sc_trace(Tf, rst, "rst");')
            ]]]*/
            //[[[end]]]
            rst.write(0);
            /*[[[cog
                cog.outl(f'{self._prog_name}.clk.bind(clk);')
                cog.outl(f'{self._prog_name}.rst.bind(rst);')
            ]]]*/
            //[[[end]]]

            /* Flush the simulation to start with, then run if
            the program hasn't already terminated.
            If a timeout argument has been passed in then run for that many
            nanoseconds. */
            rst.write(1);
            sc_start(5, SC_NS);
            rst.write(0);
            try {
                if (!sc_end_of_simulation_invoked()) {
                    if (argc > 1) sc_start(atoi(argv[1]), SC_NS);
                    else sc_start();
                }
                cout << "exiting on timeout" << endl;
                sc_close_vcd_trace_file(Tf);
            } catch (...) {
                cout << "exiting on error" << endl;
                sc_close_vcd_trace_file(Tf);
                throw;
            }

            std::cout << "exiting normally" << std::endl;
            return 0;
        }
        """
        with open(path.join(self.tempdir, f"{self._prog_name}.cpp"), "wb") as main_path:
            main_path.write(cogify(main_tmpl, globals=locals()))
        return True

    def _get_main_cpp(self, after):
        """Gets the main runtime's C++ file.

        Parameters
        ----------
        after : Coroutine
            Process which builds the C++ file.

        Returns
        -------
        BuildArtifact
            The C++ code containing :cpp:func:`sc_main`.
        """
        return BuildArtifact(f"{self._prog_name}.cpp", self, after=after)

    async def _make_top(self):
        """Makes the top module, where the wiring is defined.
        The top header also contains all node module definitions
        and the converters from Python to Migen nodes.

        Returns
        -------
        bool
            Returns True when completed.
        """
        top_tmpl = """
        /*[[[cog
            cog.outl(f'#ifndef __{self._prog_name.upper()}__')
            cog.outl(f'#define __{self._prog_name.upper()}__')
        ]]]*/
        //[[[end]]]
        #include <systemc>
        using namespace sc_core;
        using namespace sc_dt;

        // TODO: Find better way of handling trace files
        sc_trace_file *Tf;

        // Converts clock signals to bit vectors for Migen nodes
        SC_MODULE(ClkToBV) {
            sc_in<bool> clk;
            sc_out<sc_bv<1>> clkout;

            void run() {
                clkout.write(clk.read());
            }

            SC_CTOR(ClkToBV) {
                SC_METHOD(run);
                sensitive << clk;
            }
        };

        // Adaptor for going from Python to Migen
        template <class T> SC_MODULE(PythonToMigen) {
            sc_in<bool> clk;
            sc_out<T> migen_data_out;
            sc_out<sc_bv<1>> migen_valid_out;
            sc_in<sc_bv<1>> migen_ready_in;
            sc_fifo<T>* py_in;

            void run() {
                if (migen_ready_in.read() == 1) {
                    T val;
                    if (py_in->nb_read(val)) {
                        migen_data_out.write(val);
                        migen_valid_out.write(1);
                    } else {
                        migen_valid_out.write(0);
                    }
                }
            }

            SC_CTOR(PythonToMigen) {
                SC_METHOD(run);
                sensitive << clk.pos();
            }
        };

        // Adaptor for going from Migen to Python
        template <class T> SC_MODULE(MigenToPython) {
            sc_in<bool> clk;
            sc_in<T> migen_in;
            sc_in<sc_bv<1>> migen_valid_in;
            sc_out<sc_bv<1>> migen_ready_out;
            sc_fifo<T>* py_out;

            void run() {
                while (true) {
                    wait();
                    if (py_out->num_free() > 0) {
                        migen_ready_out.write(1);
                        if (migen_valid_in.read() == 1) {
                            py_out->write(migen_in.read());
                        }
                    } else {
                        migen_ready_out.write(0);
                    }
                }
            }

            SC_CTOR(MigenToPython) {
                SC_THREAD(run);
                sensitive << clk.pos();
            }
        };

        /*[[[cog
            cog.outl(f'SC_MODULE({self._get_top_name()}){{')
        ]]]*/
        //[[[end]]]

        // Python nodes to Python nodes just need a queue
        /*[[[cog
            for wire in self._py_to_py:
                name = self.get_wire_name(wire)
                cog.outl(f'sc_fifo<{PythonatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name};')
        ]]]*/
        //[[[end]]]

        // Migen nodes to Migen nodes need data wires,
        // as well as valid and ready signals
        /*[[[cog
            for wire in self._migen_to_migen:
                name = self.get_wire_name(wire)
                cog.outl(f'sc_signal<{VerilatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name}_data;')
                cog.outl(f'sc_signal<sc_bv<1>> {name}_valid;')
                cog.outl(f'sc_signal<sc_bv<1>> {name}_ready;')
        ]]]*/
        //[[[end]]]

        // Python to Migen need an adaptor which maps a queue
        // to data, valid and ready signals.
        /*[[[cog
            for wire in self._py_to_migen:
                name = self.get_wire_name(wire)
                cog.outl(f'sc_fifo<{PythonatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name}_py_out;')
                cog.outl(f'sc_signal<{VerilatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name}_migen_in;')
                cog.outl(f'sc_signal<sc_bv<1>> {name}_in_valid;')
                cog.outl(f'sc_signal<sc_bv<1>> {name}_in_ready;')
                cog.outl(f'PythonToMigen<{VerilatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name}_adaptor;')
        ]]]*/
        //[[[end]]]

        // Migen to Python need an adaptor which maps signals
        // to a queue
        /*[[[cog
            for wire in self._migen_to_py:
                name = self.get_wire_name(wire)
                cog.outl(f'sc_fifo<{PythonatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name}_py_in;')
                cog.outl(f'sc_signal<{VerilatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name}_migen_out;')
                cog.outl(f'sc_signal<sc_bv<1>> {name}_out_valid;')
                cog.outl(f'sc_signal<sc_bv<1>> {name}_out_ready;')
                cog.outl(f'MigenToPython<{VerilatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name}_adaptor;')
        ]]]*/
        //[[[end]]]

        // Python to template need a public queue
        /*[[[cog
            for wire in self._py_to_template:
                name = self.get_template_wire_name(wire, body="python", direction="out")
                cog.outl(f'sc_fifo<{PythonatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name};')
        ]]]*/
        //[[[end]]]

        // Template to python need a public queue
        /*[[[cog
            for wire in self._template_to_py:
                name = self.get_template_wire_name(wire, body="python", direction="in")
                cog.outl(f'sc_fifo<{PythonatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {name};')
        ]]]*/
        //[[[end]]]

        // Migen to template need public data, valid and ready signals
        /*[[[cog
            for wire in self._migen_to_template:
                data, valid, ready = self.get_template_wire_name(wire, body="migen", direction="out")
                cog.outl(f'sc_signal<{VerilatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {data};')
                cog.outl(f'sc_signal<sc_bv<1>> {valid};')
                cog.outl(f'sc_signal<sc_bv<1>> {ready};')
        ]]]*/
        //[[[end]]]

        // Template to Migen need public data, valid and ready signals
        /*[[[cog
            for wire in self._template_to_migen:
                data, valid, ready = self.get_template_wire_name(wire, body="migen", direction="in")
                cog.outl(f'sc_signal<{VerilatorEnv.as_c_type(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort].type)}> {data};')
                cog.outl(f'sc_signal<sc_bv<1>> {valid};')
                cog.outl(f'sc_signal<sc_bv<1>> {ready};')
        ]]]*/
        //[[[end]]]

        sc_in<bool> clk, rst;
        sc_signal<sc_dt::sc_bv<1>> rst_bv;

        // Node modules
        /*[[[cog
            for node in self._py_nodes:
                cog.outl(f'{PythonatorEnv.get_module_name(node)} {node.name};')
        ]]]*/
        //[[[end]]]

        /*[[[cog
            for node in self._migen_nodes:
                cog.outl(f'{VerilatorEnv.get_module_name(node)} {node.name};')
                cog.outl(f'ClkToBV {node.name}_clk;')
                cog.outl(f'sc_signal<sc_bv<1>> {node.name}_sysclk;')
        ]]]*/
        //[[[end]]]

        // Constructor first initialises queues, modules and adaptors
        /*[[[cog
            cog.outl(f'typedef {self._get_top_name()} SC_CURRENT_USER_MODULE;')
            cog.outl(f'{self._get_top_name()}(sc_module_name name, sc_trace_file *Tf):')

            init_modules = []
            for wire in self._py_to_py:
                init_modules.append(f'{self.get_wire_name(wire)}("{self.get_wire_name(wire)}")')

            for wire in self._py_to_migen:
                name = self.get_wire_name(wire)
                init_modules.append(f'{name}_py_out("{name}_py_out")')
                init_modules.append(f'{name}_adaptor("{name}_adaptor")')

            for wire in self._migen_to_py:
                name = self.get_wire_name(wire)
                init_modules.append(f'{name}_py_in("{name}_py_in")')
                init_modules.append(f'{name}_adaptor("{name}_adaptor")')

            for node in self._py_nodes:
                init_modules.append(f'{node.name}("{node.name}")')

            for node in self._migen_nodes:
                init_modules.append(f'{node.name}("{node.name}")')
                init_modules.append(f'{node.name}_clk("{node.name}_clk")')

            cog.out(",\\n".join(init_modules))
        ]]]*/
        //[[[end]]]
        {
            SC_METHOD(rstprop);
            sensitive << rst;

            // Wiring the clock to the Migen nodes
            /*[[[cog
                for node in self._migen_nodes:
                    cog.outl(f'{node.name}_clk.clk(clk);')
                    cog.outl(f'{node.name}_clk.clkout.bind({node.name}_sysclk);')
                    cog.outl(f'{node.name}.sys_clk.bind({node.name}_sysclk);')
                    cog.outl(f'sc_trace(Tf, {node.name}_sysclk, \"{node.name}_sysclk\");')
                    cog.outl(f'{node.name}.sys_rst.bind(rst_bv);')
            ]]]*/
            //[[[end]]]

            // Wiring the Python to Python nodes
            /*[[[cog
                for wire in self._py_to_py:
                    name = self.get_wire_name(wire)
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{PythonatorEnv.get_sysc_port_name(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort])} = &{name};')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{PythonatorEnv.get_sysc_port_name(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort])} = &{name};')
            ]]]*/
            //[[[end]]]

            // Wiring the Migen to Migen nodes
            /*[[[cog
                for wire in self._migen_to_migen:
                    name = self.get_wire_name(wire)
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_output(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="out")}.bind({name}_data);')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_output(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="in")}.bind({name}_data);')
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_valid(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="out")}.bind({name}_valid);')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_valid(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="in")}.bind({name}_valid);')
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_ready(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="in")}.bind({name}_ready);')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_ready(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="out")}.bind({name}_ready);')
            ]]]*/
            //[[[end]]]

            // Wiring the Python to Migen nodes
            /*[[[cog
                for wire in self._py_to_migen:
                    name = self.get_wire_name(wire)
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{PythonatorEnv.get_sysc_port_name(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort])} = &{name}_py_out;')
                    cog.outl(f'{name}_adaptor.clk(clk);')
                    cog.outl(f'{name}_adaptor.py_in = &{name}_py_out;')
                    cog.outl(f'{name}_adaptor.migen_data_out.bind({name}_migen_in);')
                    cog.outl(f'{name}_adaptor.migen_valid_out.bind({name}_in_valid);')
                    cog.outl(f'{name}_adaptor.migen_ready_in.bind({name}_in_ready);')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_output(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="in")}.bind({name}_migen_in);')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_valid(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="in")}.bind({name}_in_valid);')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_ready(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="out")}.bind({name}_in_ready);')
            ]]]*/
            //[[[end]]]

            // Wiring the Migen to Python nodes
            /*[[[cog
                for wire in self._migen_to_py:
                    name = self.get_wire_name(wire)
                    cog.outl(f'{name}_adaptor.clk(clk);')
                    cog.outl(f'{name}_adaptor.migen_in.bind({name}_migen_out);')
                    cog.outl(f'{name}_adaptor.migen_valid_in.bind({name}_out_valid);')
                    cog.outl(f'{name}_adaptor.migen_ready_out.bind({name}_out_ready);')
                    cog.outl(f'{name}_adaptor.py_out = &{self.get_wire_name(wire)}_py_in;')
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_output(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="out")}.bind({name}_migen_out);')
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_valid(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="out")}.bind({name}_out_valid);')
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_ready(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="in")}.bind({name}_out_ready);')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{PythonatorEnv.get_sysc_port_name(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort])} = &{name}_py_in;')
            ]]]*/
            //[[[end]]]

            // Wiring the Python to template nodes
            /*[[[cog
                for wire in self._py_to_template:
                    name = self.get_template_wire_name(wire, body="python", direction="out")
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{PythonatorEnv.get_sysc_port_name(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort])} = &{name};')
            ]]]*/
            //[[[end]]]

            // Wiring the template to Python nodes
            /*[[[cog
                for wire in self._template_to_py:
                    name = self.get_template_wire_name(wire, body="python", direction="in")
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{PythonatorEnv.get_sysc_port_name(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort])} = &{name};')
            ]]]*/
            //[[[end]]]

            // Wiring the Migen to template nodes
            /*[[[cog
                for wire in self._migen_to_template:
                    data, valid, ready = self.get_template_wire_name(wire, body="migen", direction="out")
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_output(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="out")}.bind({data});')
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_valid(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="out")}.bind({valid});')
                    cog.outl(f'{self._capnp_nodes[wire.srcNode].name}.{VerilatorEnv.get_port_ready(self._capnp_nodes[wire.srcNode].outPorts[wire.srcOutPort], direction="in")}.bind({ready});')
            ]]]*/
            //[[[end]]]

            // Wiring the template to Migen nodes
            /*[[[cog
                for wire in self._template_to_migen:
                    data, valid, ready = self.get_template_wire_name(wire, body="migen", direction="in")
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_output(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="in")}.bind({data});')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_valid(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="in")}.bind({valid});')
                    cog.outl(f'{self._capnp_nodes[wire.destNode].name}.{VerilatorEnv.get_port_ready(self._capnp_nodes[wire.destNode].inPorts[wire.destInPort], direction="out")}.bind({ready});')
            ]]]*/
            //[[[end]]]

            // Add tracing
            /*[[[cog
                for wire in self._py_to_py:
                    name = self.get_wire_name(wire)
                    cog.outl(f'{name}.trace(Tf);')
            ]]]*/
            //[[[end]]]

            /*[[[cog
                for wire in self._migen_to_migen:
                    name = self.get_wire_name(wire)
                    cog.outl(f'sc_trace(Tf, {name}_data, "{name}_data");')
                    cog.outl(f'sc_trace(Tf, {name}_valid, "{name}_valid");')
                    cog.outl(f'sc_trace(Tf, {name}_ready, "{name}_ready");')
            ]]]*/
            //[[[end]]]

            /*[[[cog
                for wire in self._py_to_migen:
                    name = self.get_wire_name(wire)
                    cog.outl(f'{name}_py_out.trace(Tf);')
                    cog.outl(f'sc_trace(Tf, {name}_migen_in, "{name}_migen_in");')
                    cog.outl(f'sc_trace(Tf, {name}_in_valid, "{name}_in_valid");')
                    cog.outl(f'sc_trace(Tf, {name}_in_ready, "{name}_in_ready");')
            ]]]*/
            //[[[end]]]

            /*[[[cog
                for wire in self._migen_to_py:
                    name = self.get_wire_name(wire)
                    cog.outl(f'{name}_py_in.trace(Tf);')
                    cog.outl(f'sc_trace(Tf, {name}_migen_out, "{name}_migen_out");')
                    cog.outl(f'sc_trace(Tf, {name}_out_valid, "{name}_out_valid");')
                    cog.outl(f'sc_trace(Tf, {name}_out_ready, "{name}_out_ready");')
            ]]]*/
            //[[[end]]]

            /*[[[cog
                for wire in self._py_to_template:
                    name = self.get_template_wire_name(wire, 'python', 'out')
                    cog.outl(f'{name}.trace(Tf);')
            ]]]*/
            //[[[end]]]

            /*[[[cog
                for wire in self._template_to_py:
                    name = self.get_template_wire_name(wire, 'python', 'in')
                    cog.outl(f'{name}.trace(Tf);')
            ]]]*/
            //[[[end]]]

            /*[[[cog
                for wire in self._migen_to_template:
                    data, valid, ready = self.get_template_wire_name(wire, body="migen", direction="out")
                    cog.outl(f'sc_trace(Tf, {data}, "{data}");')
                    cog.outl(f'sc_trace(Tf, {valid}, "{valid}");')
                    cog.outl(f'sc_trace(Tf, {ready}, "{ready}");')
            ]]]*/
            //[[[end]]]

            /*[[[cog
                for wire in self._template_to_migen:
                    data, valid, ready = self.get_template_wire_name(wire, body="migen", direction="in")
                    cog.outl(f'sc_trace(Tf, {data}, "{data}");')
                    cog.outl(f'sc_trace(Tf, {valid}, "{valid}");')
                    cog.outl(f'sc_trace(Tf, {ready}, "{ready}");')
            ]]]*/
            //[[[end]]]
        }

        void rstprop() {
            // Propogate reset signal to Migen nodes
            rst_bv.write(rst.read());
        }

        };

        #endif
        """

        with open(path.join(self.tempdir, f"{self._prog_name}.h"), "wb") as top_path:
            await multiple_waits([write_futures(header, top_path) for header in self._node_headers])
            top_path.write(cogify(top_tmpl, globals=dict(globals(), **locals())))
        return True

    def _get_top_h(self, after):
        """Get the top header file.

        Parameters
        ----------
        after : Coroutine
            Process which builds the header file.

        Returns
        -------
        BuildArtifact
            The top header file.
        """
        return BuildArtifact(f"{self._prog_name}.h", self, after=after)

    async def _build_main(self, after):
        """Builds the main.o binary object containing the
        :cpp:func:`sc_main` runtime

        Parameters
        ----------
        after : Coroutine
            The process which builds the main runtime's C++ file.

        Returns
        -------
        bool
            Returns True when complete.
        """
        done = await self._run_gcc(f"{self._prog_name}", after)
        return done

    def _get_main_object(self, after):
        """Get the binary object conatining sc_main.

        Parameters
        ----------
        after : Coroutine
            Process which builds the binary object.

        Returns
        -------
        BuildArtifact
            The :cpp:func:`sc_main` binary object.
        """
        return BuildArtifact(f"{self._prog_name}.o", self, after=after)

    async def _write_objects(self):
        """Writes all objects in to the temporary directory.

        Returns
        -------
        bool
            Returns True when complete.
        """
        object_paths = [open(path.join(self.tempdir, module.name), "wb") for module in self._node_objects]
        await multiple_waits([write_futures(module, object_path) for module, object_path in zip(self._node_objects, object_paths)])
        for object_path in object_paths:
            object_path.close()
        return True

    async def _make_archive(self, after):
        """Run ar to compile a .a archive containing the
        binary objects for all nodes.

        Parameters
        ----------
        after : list
            :class:`Coroutine` objects which writes the objects to
            the environment's temporary directory.

        Returns
        -------
        bool
            Returns True upon completion.
        """
        done = await self._run_ar([object_file.name for object_file in self._node_objects], after=after, name=self._prog_name)
        return done

    async def _link_objects(self, main_o, archive, after):
        """Links sc_main binary object with all other objects
        to create a complete runtime.

        Parameters
        ----------
        main_o : BuildArtifact
            The binary object containing the :cpp:func:`sc_main` runtime.
        archive : BuildArtifact
            The .a archive containing all other objects.
        after : list
            Coroutines for building the main object and archive.

        Returns
        -------
        bool
            Returns True when complete.
        """
        done = await self._link(main_o.name, archive, after=after, name=self._prog_name)
        return done

    def wiring(self, capnp_graph):
        """Wire the full graph together and return the complete build.

        Parameters
        ----------
        capnp_graph
            capnp object describing the full wiring

        Returns
        -------
        dict
            Map from strings to :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects.
            Mapping is the following, where `prog_name` is the name of
            the program:

            - "{prog_name}.h": the header file containing all node headers and the full wiring module.
            - "{prog_name}.a": .a archive containing all built objects.

            If there are no template nodes the following are also provided:

            - "{prog_name}.cpp": C++ code containing :cpp:func:`sc_main` for complete runtime
            - "{prog_name}.o" binary object for :cpp:func:`sc_main`
            - "{prog_name}" complete runtime executable.

        """
        self._get_adaptors(capnp_graph)
        make_top = self._make_top()
        top_h = self._get_top_h(make_top)
        if not self._has_templates:
            make_main = self._make_main()
            main_file = self._get_main_cpp(make_main)
            build_main = self._build_main([make_main, make_top])
            main_o = self._get_main_object(build_main)
        write_objects = self._write_objects()
        make_archive = self._make_archive([write_objects])
        main_a = self._get_archive(make_archive, name=self._prog_name)
        if not self._has_templates:
            link_objects = self._link_objects(main_o, main_a, [build_main, make_archive])
            complete_runtime = self._get_main(link_objects, name=self._prog_name)
        build_artifacts = dict()
        build_artifacts[f"{self._prog_name}.h"] = top_h
        build_artifacts[f"{self._prog_name}.a"] = main_a
        if not self._has_templates:
            build_artifacts[f"{self._prog_name}.cpp"] = main_file
            build_artifacts[f"{self._prog_name}.o"] = main_o
            build_artifacts[f"{self._prog_name}"] = complete_runtime
        return build_artifacts



import tempfile


class Environment():
    """This base class supports defining a context manager providing
    access to build tools and a isolated enviroment.

    Attributes
    ----------

    tempdir : TemporaryDirectory
        The temporary directory this environment is within.
    """

    def __init__(self):
        self._tempdirmanger = tempfile.TemporaryDirectory()
        self.tempdir = None

        self._check_env_ok() # override this function to ensure all tools are present

    # I think we intend for the user to be making the contexts, so they will always
    # be created from non-async python. If we want to enter subcontexts we will need
    # __aenter__ as well.
    def __enter__(self):
        """enter the temp dir."""
        self.tempdir = self._tempdirmanger.__enter__()
        return self

    def __exit__(self, type, value, traceback):
        self._exit_info = (type, value, traceback)

    def __del__(self):
        type, value, traceback = self._exit_info
        self._tempdirmanger.__exit__(type, value, traceback)
        self.tempdir = None

    def _check_env_ok(self):
        """Overwritten by child environments to check the Environment has
        all the necessary tools.
        """
        return True



from deltasimulator.build_tools.utils import wait
import asyncio
import os.path
import os
import traceback
import logging
log = logging.getLogger(__name__)


class BuildArtifact():
    """Wrapper object for asynchronously building a file.

    Parameters
    ----------
    name : str
        The name of the artifact. Typically the file name.
    env : Optional[Environment]
        The environment that the artifact was built within. This is used
        to know which temporary directory the artifact is within.
        By default `None`.
    path : Optional[str]
        The file path within the temp directory, by default "."
    after : Optional[Coroutine]
        An asynchronous coroutine which needs to complete before the artifact
        has been built
    data : Union[bytes, bytearray]
        If the artefact's content is already available then it can be
        provided as-is rather than needing to build it in an environment.


    .. warning::
        If `env` is `None` then `data` must not be `None`,
        otherwise a :exc:`ValueError` will be raised.

    .. warning::
        If `data` is provided then it must be bytes or byteslike,
        otherwise a :exc:`TypeError` will be raised.
    """

    def __init__(self, name: str, env=None, path: str = ".", after=None, data=None):
        self.name = name
        self.path = path

        if data is not None: # literal provided
            if not isinstance(data, (bytes, bytearray)):
                raise TypeError(
                    f"If data is provided it must be bytes or byteslike, was {type(data)}")
            self._data = data  # if we have bytes, use them
            self.env = None
            self._flag = None
        else:  # no data - need to look it up in an env
            self._data = None
            if env is None:
                raise ValueError(
                    "If data is not provided, we must know the env to look the file up within")
            self.env = env
            if env.tempdir is None:
                raise RuntimeError(
                    f"env {env} does not have a tempdir for path lookup")
            # flag is a coro that when completed, indicates
            # the file is available to be read
            self._flag = after

    @property
    async def data(self):
        """The content of the :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>`.
        If data was not provided when constructed then the content is
        retrieved from the build environment's temporary directory,
        building the file if needed by waiting for the :class:`Coroutine`
        to complete.

        Returns
        -------
        bytes
            The content of the artifact.

        Raises
        ------
        RuntimeError
            If there is no data and no :class:`Environment` is provided
            then an error will be raised.
        FileNotFoundError
            Raised if the file wrapped by the BuildArtifact is not found.
            This might be because there was an error during the build process
            which resulted in the file not being created.
        """
        if self._data is not None:
            return self._data
        else:
            if self._flag is not None:
                await wait(self._flag)
                log.debug(f"flag {self._flag} resolved")
            if self.env.tempdir is None:
                raise RuntimeError(
                    f"env {self.env} does not have a tempdir for path lookup")
            try:
                path = os.path.join(self.env.tempdir, self.path, self.name)
                with open(path, "rb") as file:
                    self._data = file.read()
                    return self._data
            except FileNotFoundError as e:
                # try and be helpful
                log.error(
                    f"file lookup failed: dir contents {list(os.walk(path))}, {e}")
                log.error(
                    f"tmp dir root contents {os.listdir(self.env.tempdir)}, {e}")
                log.error(
                    f"you probably forgot a after call - we expect to be after {self._flag}, {e}")
                log.error(f"Full traceback: {traceback.format_exc()}, {e}")

                # we currently hold a ref to env, that will preserve the tempdirs
                # that should hold this file. We can be helpful and drop to a shell
                # so the user can investigate the problem
                if False:
                    log.error(
                        "Dropping to a shell, press CTRL+d to exit and continue")
                    os.chdir(self.env.tempdir)
                    import pty
                    pty.spawn("/bin/bash")

                # after the shell is done, the user has investigated the problem
                # we should just terminate the process
                raise


class BuildArtifactSet():
    """Wrapper object for asynchronously building a set of files.

    A :class:`BuildArtifactSet` has the same API as a :class:`dict` object in Python.
    """

    def __init__(self):
        self._store = dict()

    async def __getitem__(self, key, type=None):
        print(self._store)
        return await self._store.__getitem__(key)

    async def __setitem__(self, key, value):
        # value = await asyncio.wait_for(value, timeout=None)
        self._store.__setitem__(key, value)


async def write_futures(object: BuildArtifact, output_file):
    """Asynchronously write the contents of the BuildArtifact to file. 
    Use if already within an async coroutine.

    Parameters
    ----------
    object : BuildArtifact
        The artifact to be written.
    output_file
        The file to be written to.
    """
    try:
        data = await asyncio.wait_for(object.data, timeout=None)
    except AttributeError:
        log.error(f"failed to write {object}")
        raise
    output_file.write(data)


def write(object: BuildArtifact, output_file):
    """Asynchronously write the contents of the BuildArtifact to file. 

    Parameters
    ----------
    object : BuildArtifact
        The artifact to be written.
    output_file
        The file to be written to.


    .. warning::
        Will raise an error if called within an asynchronous process.
        Use :meth:`write_futures` instead.
    """

    asyncio.run(write_futures(object, output_file))



import asyncio
import unittest
from os import path
import re

from deltalanguage.data_types import DOptional
from deltalanguage.wiring import DeltaGraph, Interactive, PyInteractiveNode
from deltalanguage.runtime import serialize_graph
from deltasimulator.build_tools.environments import PythonatorEnv

from test._utils import add, const_exit, print_then_exit

@Interactive(in_params={"num": int, "val": int, "opt": DOptional(int)}, out_type=int, name="pythonate_interactive")
def interactive_func(node: PyInteractiveNode):
    while(True):
        num = node.receive()["num"]
        print (f"received num = {num}")
        opt = node.receive()["opt"]
        if opt:
            print (f"received opt={opt}")
        val = node.receive()["val"]
        node.send(num + val + 1)

class TestPythonator(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestPythonator, self).__init__(*args, **kwargs)
        self.build_artifacts = None

    async def assert_build_correct(self, prefix):
        print (f"checking {prefix} artifacts")
        built = {}
        built["cpp"] = await asyncio.wait_for(self.build_artifacts["cpp"].data,
                                              timeout=None)
        built["h"] = await asyncio.wait_for(self.build_artifacts["h"].data,
                                            timeout=None)
        if "py" in self.build_artifacts:
            built["py"] = await asyncio.wait_for(self.build_artifacts["py"].data,
                                                 timeout=None)
        built["o"] = await asyncio.wait_for(self.build_artifacts["o"].data,
                                            timeout=None)
        maxDiff = self.maxDiff
        self.maxDiff = None
        with open(path.join("test", "data", "pythonate", f"{prefix}_pythonate_cpp.out"), "rb") as cpp_file:
            gen = built["cpp"].decode("utf-8")
            ref = cpp_file.read().decode("utf-8")
            self.assertMultiLineEqual(gen, ref, f"{prefix}_pythonate_cpp.out does not match built file")
        with open(path.join("test", "data", "pythonate", f"{prefix}_pythonate_h.out"), "rb") as h_file:
            ref = h_file.read().decode("utf-8")
            gen = built["h"].decode("utf-8")
            self.assertMultiLineEqual(gen, ref, f"{prefix}_pythonate_h.out does not match built file")
        with open(path.join("test", "data", "pythonate", f"{prefix}_pythonate_py.out"), "rb") as py_file:
            # Remove pickled portion - we can't guarantee the same result with unittest vs normal generation
            ref = py_file.read().decode("utf-8")
            gen = built["py"].decode("utf-8")
            filt = r"loads\(.*\)"
            self.assertMultiLineEqual(re.sub(filt,'loads(...)',gen, flags=re.MULTILINE),
                                      re.sub(filt,'loads(...)',ref, flags=re.MULTILINE),
                                      f"{prefix}_pythonate_py.out does not match built file")
        self.maxDiff = maxDiff

    def test_pythonate(self):
        """Build artifacts and check the compiled code is correct."""
        with DeltaGraph() as test_graph:
            print_then_exit(n=interactive_func.call(num=add(a=2, b=3), val=4, opt=4))

        _, program = serialize_graph(test_graph)
        for node in program.nodes:
            if "add" in node.name:
                with PythonatorEnv(program.bodies) as env:
                    self.build_artifacts = env.pythonate(node)
                asyncio.run(self.assert_build_correct("add"))
            elif "interactive" in node.name:
                with PythonatorEnv(program.bodies) as env:
                    self.build_artifacts = env.pythonate(node)
                asyncio.run(self.assert_build_correct("interactive"))

    def test_const_exit(self):
        with DeltaGraph() as test_graph:
            const_exit(a=5)

        _, program = serialize_graph(test_graph)
        with PythonatorEnv(program.bodies) as env:
            self.build_artifacts = env.pythonate(program.nodes[1])
        with self.assertRaises(ValueError):
            asyncio.run(self.build_artifacts["cpp"].data)

    def tearDown(self):
        DeltaGraph.clean_stack()


if __name__ == "__main__":
    unittest.main()



import asyncio
from itertools import zip_longest
import logging
import os
import re
import subprocess as sp
import traceback

import dill

# from typing import TYPE_CHECKING
# if TYPE_CHECKING:
from deltasimulator.build_tools import Environment, BuildArtifact
from deltasimulator.build_tools.utils import wait

log = logging.getLogger(__name__)


def grouper(iterable, n, fillvalue=None):
    """Collect data into fixed-length chunks or blocks.
    Taken from :py:mod:`itertools` recipes.
    """
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


class VerilatorEnv(Environment):
    """Environment for running Verilator to generate SystemC modules of
    Migen nodes.

    .. note::
        Migen nodes communicate using three signals:

        - a data signal which the data is sent along,

        - a valid signal to indicate that there is data to be sent,

        - a ready signal to indicate that the node is ready to receive data.

    Attributes
    ----------

    env_ok
        Set once to ensure the environment is okay to run.


    .. warning::
        Verilator 4.0 or greater is required for the environment to run.

    """

    env_ok = None  # only check once per Env (performance)

    def _check_env_ok(self):
        """Checks that the environment is able to run.

        Returns
        -------
        bool
            True if the environment is okay.

        Raises
        ------
        RuntimeError
            Raised if Verilator version is earlier than 4.0 or if any other
            exception is raised while checking the version.
        """
        if self.env_ok is not None:
            if not self.env_ok:
                raise RuntimeError("env not ok from previous check")
            else:
                return self.env_ok
        else:
            # do real checking
            try:
                output = sp.run(["verilator", "--version"],
                                capture_output=True,
                                check=True)
                version = output.stdout.split(b" ")[1]
                if float(version) < 4.0:
                    raise RuntimeError(
                        f"verilator version too old, need >4 got {version}")
            except Exception as ex:
                traceback.print_exc()
                raise RuntimeError(
                    f"failed to check enviroment, got error {ex}") from ex
            else:
                self.env_ok = True

    async def _run_verilator(self, top_v: BuildArtifact):
        """This process builds a top.v into a cpp file. First step.
        -${DRUN} verilator -CFLAGS "${CPPFLAGS}" --sc $< \
                           --Mdir build/verilated/  \
                               --top-module $(basename $(notdir $<)) \
                                -I$(dir $<) \
                                -Wno-lint

        Parameters
        ----------
        top_v: BuildArtifact
            The Verilog code for the module.


        .. warning::
            `top_v` is expected to have the same name as its module.
            If this is not the case then the module name will be replaced.

        .. warning::
            ROM init blocks are not supported by Verilator. In the Verilog
            contains an init block then the init files are written separately.
        """
        data = await asyncio.wait_for(top_v.data, timeout=None)
        mod_name = top_v.name.split(".")[0].encode("utf8")

        # verify the module name is correct
        name_regex = b'module\\s+([a-zA-Z0-9\\_]+)\\('
        mod_internal_name = re.search(name_regex, data)[1]  # first subgroup
        if mod_internal_name != mod_name:
            log.warning(
                "WARNING: migen node name doesn't match its module name"
                f"This is not supported! expected: {mod_name},"
                f"got: {mod_internal_name}"
            )
            data = re.sub(name_regex, b"module " +
                          mod_name+b"(", data)

        # look for inline ROM init files!
        # verilator does not suppport this, so we need to write them seperately
        module_regex = b"([a-zA-Z0-9\\_]+)\\.init:\n"
        matches = re.split(module_regex, data)
        if len(matches) > 1:
            for rom_name, rom_data in grouper(matches[1:], 2):
                log.info(
                    "INFO: found a ROM init block called"
                    f"{rom_name} in {mod_name}"
                    "This is not supported by Verilator."
                    "Extracting the data by hand."
                )
                full_string = (
                    f'{rom_name.decode("utf-8")}.init:'
                    f'\n{rom_data.decode("utf-8")}'
                ).encode("utf-8")
                # remove the ROM data from the file
                data = data.replace(full_string, b'')
                init_file = f"{rom_name.decode('utf-8')}.init"
                fpath = os.path.join(self.tempdir, init_file)
                with open(fpath, "wb") as f:
                    f.write(rom_data)
                log.info(f"wrote {len(rom_data)} bytes to {rom_name}.init")
                print(rom_data)

        with open(os.path.join(self.tempdir, top_v.name), "wb") as f:
            f.write(data)

        # run the build step!
        log.info("running build")
        cmd = ["verilator",
               "--sc", top_v.name,  # build a module for SystemC
               "--top-module", top_v.name.split(".")[0],
               "-CFLAGS", "-std=c++17 -DSC_CPLUSPLUS=201703L \
                   -D_FORTIFY_SOURCE=2 -faligned-new -fstack-protector \
                        -Wall -Werror -g -grecord-gcc-switches -O2",
               "-Wno-fatal", "-Wno-combdly", "-Wno-unoptflat",
               "-Wno-width", "-Wno-initialdly",
               # ensure consistant port types by making everything a bv
               "--pins-bv", "1"
               ]

        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=self.tempdir)

        stdout, stderr = await proc.communicate()
        log.info(stdout.decode("utf8"))
        log.warning(stderr.decode("utf8"))
        return True

    def _get_cpp(self, top_v, after):
        """Returns the C++ file generated by Verilator.

        Parameters
        ----------
        top_v : BuildArtifact
            The Verilog code.
        after : Coroutine
            Process to wait for before the C++ code is ready.

        Returns
        -------
        BuildArtifact
            The C++ code generated.
        """
        mod_name = top_v.name.split(".")[0]
        cpp_name = "V"+mod_name+".cpp"
        cpp_path = "obj_dir"
        return BuildArtifact(
            name=cpp_name,
            path=cpp_path,
            after=after,
            env=self
        )

    def _get_h(self, top_v, after):
        """Get the header file generated by Verilator.

        Parameters
        ----------
        top_v : BuildArtifact
            The Verilog code.
        after : Coroutine
            Process to wait for before the header file is generated.

        Returns
        -------
        BuildArtifact
            The generated header file.
        """
        mod_name = top_v.name.split(".")[0]
        h_name = "V"+mod_name+".h"
        h_path = "obj_dir"
        return BuildArtifact(
            name=h_name,
            path=h_path,
            after=after,
            env=self
        )

    def _get_rom_init(self, top_v, after):
        """Check if there are any ROM init files and if so return them.

        Parameters
        ----------
        top_v : BuildArtifact
            The Verilog code.
        after : Coroutine
            Process which builds the Verilated code.

        Returns
        -------
        list
            :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>`
            objects for the ROM init files.
        """
        # Get the Verilog code and search for ROM init code by regex.
        # TODO: Find a nicer way of getting the init file name.
        data = asyncio.run(top_v.data)
        module_regex = b"([a-zA-Z0-9\\_]+)\\.init:\n"
        inits = []
        for match in re.finditer(module_regex, data):
            rom_name = match[0][:-2].decode("utf8")
            inits.append(BuildArtifact(
                name=rom_name,
                env=self,
                after=after
            ))
        return inits

    async def _build_objects(self, top_v, h, after):
        """Once C++ and header files are generated, use make to compile them
        into a .a archive.

        Parameters
        ----------
        top_v : BuildArtifact
            The Verilog code. Only the file name is used here.
        h : BuildArtifact
            The header file generated.
        after : Coroutine
            Code can be compiled after this :class:`Coroutine` has finished.

        Returns
        -------
        bool
            Returns True upon completion.
        """
        # given _build_Vcpp has run in tempdir, use make to build the
        # required object files.
        # we could possibly get a little more parallelism here by doing each
        # target individually.

        # we don't need the file, just the name
        # import code; from pprint import pprint;
        # vp=lambda obj: pprint(vars(obj)); code.interact(local=locals())

        # ensure .cpp is present by waiting for flag
        await wait(after)
        log.debug(
            f"verilator env before object build {os.listdir(self.tempdir)}")
        mod_name = top_v.name.split(".")[0]

        proc = await asyncio.create_subprocess_exec(
            "make", "-C", "obj_dir/", "-j", "-f", "V" +
            mod_name+".mk", "V"+mod_name+"__ALL.a",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=self.tempdir)

        stdout, stderr = await proc.communicate()
        log.info(stdout.decode("utf8"))
        if stderr:
            log.warning(stderr.decode("utf8"))
        log.info(
            f"{os.listdir(self.tempdir)},"
            f"{os.listdir(os.path.join(self.tempdir, 'obj_dir'))}"
        )
        return True

    async def _build_verilated_o(self, top_v, after):
        """Build the verilated.o binary file.
        This is required when linking nodes which were built using Verilator.

        Parameters
        ----------
        top_v : BuildArtifact
            The Verilog code. Only the file name is used here.
        after : Coroutine
            :class:`Coroutine` that needs to finish before the verilated.o
            file can be built.

        Returns
        -------
        bool
            Returns True upon completion.
        """
        await wait(after)
        mod_name = top_v.name.split(".")[0]

        proc = await asyncio.create_subprocess_exec(
            "make", "-C", "obj_dir/", "-j", "-f",
            "V"+mod_name+".mk", "verilated.o",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=self.tempdir)

        stdout, stderr = await proc.communicate()
        log.info(stdout.decode("utf8"))
        if stderr:
            log.warning(stderr.decode("utf8"))
        return True

    def _get_verilated_o(self, after):
        """Once verilated.o is built, return the file.

        Parameters
        ----------
        after : Coroutine
            :class:`Coroutine` which is building the verilated.o file.

        Returns
        -------
        BuildArtifact
            The verilated.o binary object.
        """
        return BuildArtifact(
            name="verilated.o",
            path="obj_dir/",
            env=self,
            after=after
        )

    def _get_ALL_a(self, top_v, after):
        """Once the binary objects have been built return them in a .a archive.

        Parameters
        ----------
        top_v : BuildArtifact
            Verilog code file. Only used for the file name.
        after : Coroutine
            :class:`Coroutine` which builds the binary objects.

        Returns
        -------
        BuildArtifact
            The binary objects in a .a archive.
        """
        mod_name = top_v.name.split(".")[0]
        h_name = "V"+mod_name+"__ALL.a"
        h_path = "obj_dir"
        return BuildArtifact(
            name=h_name,
            path=h_path,
            env=self,
            after=after
        )

    # def _get_verilated_sc_h(self):
    #     this is just in the same dir as verilator.
    #     path = sp.run(["which", "verilator"],
    #       capture_output=true, check=True).stdout
    #     return BuildArtifact(name="verilated_sc.h",
    #       path="obj", env=self, after=None)
    # TODO: use this rather than letting cxxenv look for headers

    def verilate(self, top_v: BuildArtifact):
        """Runs Verilator and builds the binary objects.

        Parameters
        ----------
        top_v : BuildArtifact
            The Verilog code.

        Returns
        -------
        dict
            A map of strings to :class:
            `BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects:

            - "cpp": the C++ file
            - "h": the header file
            - "init": list of ROM init files
            - "ALL.a": the .a archive containing the binary objects
            - "verilated.o": the verilated.o binary object
        """
        # fist step: generate a simulation module
        verilated = self._run_verilator(top_v)
        cpp = self._get_cpp(top_v, after=verilated)
        h = self._get_h(top_v, after=verilated)

        # second step: compile the top.cpp into a shared object
        binary = self._build_objects(top_v, h, after=verilated)
        all_a = self._get_ALL_a(top_v, after=binary)

        # return a set of build artifacts
        built_artifacts = dict()
        built_artifacts["cpp"] = cpp
        built_artifacts["h"] = h
        built_artifacts["init"] = self._get_rom_init(
            top_v=top_v, after=verilated)
        built_artifacts["ALL.a"] = all_a
        built_artifacts["verilated.o"] = self._get_verilated_o(
            after=self._build_verilated_o(top_v, after=verilated))
        # built_artifacts["verilated_sc.h"] =
        #   self._get_verilated_sc_h(after=binary)
        return built_artifacts

    @staticmethod
    def as_c_type(df_type):
        """Gives the SystemC equivalent of the given type.

        Parameters
        ----------
        df_type : bytes
            The dill serialisation of the DeltaType.

        Returns
        -------
        str
            String of the SystemC equivalent.


        .. note::
            PyMigenNode is a subclass of PythonNode and
            uses the same data_types for I/O ports.
            For type format see :meth:`PythonatorEnv.as_c_type`.
        """
        return f"sc_bv<{dill.loads(df_type).size}>"

    @staticmethod
    def get_sysc_port_name(port, direction="out") -> tuple:
        """Get the wire names for a Migen node's ports.

        Parameters
        ----------
        port
            capnp object describing the port.
        direction : Optional[str]
            The direction of the port, by default "out"

        Returns
        -------
        tuple
            Names for the data, valid and ready wires.
            Format for wire names is `{port.name}_{direction}_{type}`,
            where `port_name` is the name of the port,
            `direction` is "out" or "in" and `type` is "data", "valid"
            or "ready".


        .. note::
            The direction of the ready wire is the opposite of the direction
            of the data and valid wires.
        """
        if direction == "out":
            return (
                f"{port.name}_out_data",
                f"{port.name}_out_valid",
                f"{port.name}_in_ready"
            )
        elif direction == "in":
            return (
                f"{port.name}_in_data",
                f"{port.name}_in_valid",
                f"{port.name}_out_ready"
            )

    @staticmethod
    def get_port_output(port, direction="out"):
        """Get the name of the port's data wire.

        Parameters
        ----------
        port
            capnp object describing the port.
        direction : Optional[str]
            The direction of the port, by default "out"

        Returns
        -------
        str
            Name for the port's data wire.
            See :meth:`VerilatorEnv.get_sysc_port_name` for the naming format.
        """
        if direction == "out":
            return f"{port.name}_out_data"
        elif direction == "in":
            return f"{port.name}_in_data"

    @staticmethod
    def get_port_valid(port, direction="out"):
        """Get the name of the port's valid wire.

        Parameters
        ----------
        port
            capnp object describing the port.
        direction : Optional[str]
            The direction of the port, by default "out"

        Returns
        -------
        str
            Name for the port's valid wire.
            See :meth:`VerilatorEnv.get_sysc_port_name` for the naming format.
        """
        if direction == "out":
            return f"{port.name}_out_valid"
        elif direction == "in":
            return f"{port.name}_in_valid"

    @staticmethod
    def get_port_ready(port, direction="out"):
        """Get the name of the port's ready wire.

        Parameters
        ----------
        port
            capnp object describing the port.
        direction : Optional[str]
            The direction of the port, by default "out"

        Returns
        -------
        str
            Name for the port's ready wire.
            See :meth:`VerilatorEnv.get_sysc_port_name` for the naming format.
        """
        if direction == "out":
            return f"{port.name}_in_ready"
        elif direction == "in":
            return f"{port.name}_out_ready"

    @staticmethod
    def get_module_name(top_p):
        """Get the name of the SystemC module for a Migen node.

        Parameters
        ----------
        top_p
            capnp object describing the Migen node.

        Returns
        -------
        str
            The name of the Migen node's SystemC module.
            The format is `V{node_name}` where `node_name`
            is the name of the node.
        """
        return "V"+top_p.name



import asyncio
from os import path
import subprocess
import sys
import zipfile

from deltasimulator.build_tools import BuildArtifact, write
from deltasimulator.build_tools.environments import (PythonatorEnv,
                                                     VerilatorEnv,
                                                     WiringEnv,
                                                     CPPEnv,
                                                     HostEnv)
from capnp.lib.capnp import _DynamicStructBuilder


def generate_wiring(program: _DynamicStructBuilder) -> dict:
    """Creates the wiring of the nodes defined in a program.

    Parameters
    ----------
    program: _DynamicStructBuilder
        A Deltaflow serialized graph
    Returns
    -------
    node_bodies: list
        the bodies of the extracted nodes
    node_inits: list
        all the rom files found in the migen nodes, extracted as strings.
        This solves an incompatibility between some migen generated outputs
        and verilator.
    wiring: dict
        the wiring of the graph. This can be used to generate a SystemC top
        level to wire the graph
    """

    node_headers = []
    node_bodies = []
    node_modules = []
    node_objects = []
    node_inits = []
    verilated_o = None
    for node in program.nodes:
        if node.body != -1:
            # If node.body is -1 then it is a template node
            # and cannot be pythonated
            which_body = program.bodies[node.body].which()

            if which_body in ['python', 'interactive']:
                with PythonatorEnv(program.bodies) as env:
                    build_artifacts = env.pythonate(node)
                    node_headers.append(build_artifacts["h"])
                    if "py" in build_artifacts:
                        node_bodies.append(build_artifacts["py"])
                    node_modules.append(build_artifacts["cpp"])
                    node_objects.append(build_artifacts["o"])

            elif which_body == 'migen':
                # This part is adopted from initial-example:
                top_v = BuildArtifact(name=f"{node.name}.v",
                                      data=program.bodies[node.body].migen.verilog.encode("utf8"))

                with VerilatorEnv() as env:
                    build_artifacts = env.verilate(top_v)

                node_headers.append(build_artifacts["h"])
                node_modules.append(build_artifacts["cpp"])
                node_objects.append(build_artifacts["ALL.a"])
                node_inits += build_artifacts["init"]
                if not verilated_o:
                    verilated_o = build_artifacts["verilated.o"]

    with WiringEnv(program.nodes,
                   program.bodies,
                   node_headers,
                   node_objects,
                   verilated_o,
                   program.name) as env:
        wiring = env.wiring(program.graph)

    return node_bodies, node_inits, wiring


async def _wait_for_build(wiring: dict):
    """Waits for all the objects associated with the wiring to be
    ready.

    Parameters
    ----------
    wiring
        A wired graph
    """

    for comp in wiring:
        _ = await asyncio.wait_for(wiring[comp].data, timeout=None)


def _copy_artifacts(main, inits, node_bodies, destination):
    """Copies all the required artifacts into a new destination

    Parameters
    ----------
    main
        the main file (main.cpp) for the graph
    inits
        the memory configuration files
    node_bodies
        the content of the nodes
    destination
        the folder to copy the artifacts into. Note: created if it does
        not exists
    """

    for init in inits:
        with open(path.join(destination, init.name), "wb") as f:
            write(init, f)

    for body in node_bodies:
        with open(path.join(destination, body.name), "wb") as f:
            write(body, f)

    with open(path.join(destination, "main"), "wb") as f:
        write(main, f)


def _compile_and_link(program_name: str, wiring: list, main_cpp: str):
    """Compiles and link the graph together with a provided top
    level file

    Parameters
    ----------
    program_name: str
        the name of the program to generate
    wiring: list
        a wired graph generated via cog
    main_cpp: str
        the top level main file
    """

    asyncio.run(_wait_for_build(wiring))
    _main_h = wiring[program_name + ".h"]
    _main_a = wiring[program_name + ".a"]
    # Converting the main.cpp into a BuildArtifact
    with HostEnv(dir=path.dirname(main_cpp)) as env:
        _main_cpp = BuildArtifact(name=path.basename(main_cpp), env=env)

    with CPPEnv() as cpp:
        main = cpp.compile_and_link(
            [_main_h],
            [_main_a],
            _main_cpp
        )
    return main


def build_graph(program: _DynamicStructBuilder, main_cpp: str, build_dir: str):
    """ Generates an executable to be stored in a build
    directory.

    Parameters
    ----------
    program: _DynamicStructBuilder
        The Deltaflow program to be converted into SystemC.
    main_cpp: str
        SystemC top level file that starts the SystemC simulation and
        defines an implementation for eventual templatedNodes.
    build_dir: str
        The target directory in which to store the output.

    Examples
    --------
    The *main.cpp* (other filenames are allowed) should at least contain the
    following:

    .. code-block:: c++

        #include <systemc>
        #include <Python.h>
        #include "dut.h"
        using namespace sc_dt;
        int sc_main(__attribute__(int argc, char** argv) {
            Py_UnbufferedStdioFlag = 1;
            Py_Initialize();
            sc_trace_file *Tf = nullptr;
            sc_clock clk("clk", sc_time(1, SC_NS));
            sc_signal<bool> rst;
            // Dut is the name you
            Dut dut("Dut", Tf);
            dut.clk.bind(clk);
            dut.rst.bind(rst);
            rst.write(0);
            sc_start(1000, SC_NS);
            return 0;
        }


    With the associate Python code:

    .. code-block:: python

        from deltalanguage.runtime import serialize_graph
        from deltasimulator.lib import build_graph
        ...
        _, program = serialize_graph(graph, name="dut")
        build_graph(program, main_cpp="main.cpp",
             build_dir="/workdir/build")
        ...

    .. todo::

        Setting stdio/stderr to not buffer is required because `Py_Finalize`
        can cause some memory errors (with eg the Qiskit HAL test).
        Is there a fix for this?

    """

    if len(program.requirements) > 0:
        req_path = path.join(build_dir, "requirements.txt")
        with open(req_path, "w") as req_txt:
            req_txt.write("\n".join(program.requirements))
        try:
            subprocess.run([sys.executable,
                            '-m',
                            'pip',
                            'install',
                            '-r',
                            req_path],
                           capture_output=True,
                           check=True)
        except subprocess.CalledProcessError as e:
            raise RuntimeError("Error with installing required dependencies:",
                               e.output) from e
    node_bodies, node_inits, wiring = generate_wiring(program)
    main = _compile_and_link(program.name, wiring, main_cpp)
    _copy_artifacts(main, node_inits, node_bodies, build_dir)
    if program.files != b'':
        zip_name = path.join(build_dir, "df_zip.zip")
        with open(zip_name, "wb") as zip_file:
            zip_file.write(program.files)
        df_zip = zipfile.ZipFile(zip_name, "r")
        if df_zip.testzip() is None:
            df_zip.extractall(build_dir)
        else:
            raise RuntimeError("Corrupted supporting files")



import unittest
from deltasimulator.build_tools.cogify import cogify


class TestCogify(unittest.TestCase):

    def test_no_transformation(self):
        teststr = "hello, \n\0"
        result = cogify(teststr)
        self.assertEqual(result, teststr.encode("utf-8"))

    def test_block(self):
        teststr = """
        #[[[cog
        #   for i in range(5):
        #       cog.outl(f"{i}")
        #]]]
        #[[[end]]]
        """
        result = cogify(teststr)
        resstr = b"\n0\n1\n2\n3\n4\n"
        self.assertEqual(resstr, result)

    def test_strip_tabs(self):
        teststr = """
        hello
        world"""
        result = cogify(teststr)
        resstr = b"\nhello\nworld"
        self.assertEqual(resstr, result)


if __name__ == "__main__":
    unittest.main()



from deltasimulator.build_tools import Environment

class HostEnv(Environment):
    """A shim enviroment representing the host filesystem.

    Parameters
    ----------
    dir : str
        The path to the host directory.
    """

    def __init__(self, dir):
        self.tempdir = dir

    def __enter__(self):
        # nothing to do, as we don't need to isolate the host from itself.
        return self

    def __exit__(self, type, value, traceback):
        self._exit_info = (type, value, traceback)

    def __del__(self):
        self.tempdir = None



import asyncio
import os
import sysconfig
import shlex
import logging

from find_libpython import find_libpython
from collections.abc import Iterable

from deltasimulator.build_tools import BuildArtifact, Environment
from deltasimulator.build_tools.utils import multiple_waits


log = logging.getLogger(__name__)


class CPPEnv(Environment):
    """Environment for running C++-related commands such as g++ and ar.

    .. note::
        Any stdout from these commands is written to :meth:`logging.debug`.
        Any stderr from them is written to :meth:`logging.warning`.
    """

    CFLAGS = ["-std=c++17",
              "-DSC_CPLUSPLUS=201703L",
              "-D_FORTIFY_SOURCE=2",
              "-faligned-new",
              "-fstack-protector",
              "-Wall", "-Wextra",
              # every warning but we use a lot of templated code so unused is common
              "-Wno-unused-variable", "-Wno-unused-parameter",
              "-Werror",
              "-g",
              "-grecord-gcc-switches",
              "-O2",
              "-fPIE"]

    async def _run_gcc(self, module_name: str, after):
        """Runs g++ on given module to compile to an object.

        Parameters
        ----------
        module_name : str
            The name of the module. Note that ".cpp" is appended by _run_gcc.
            We might change this behaviour if we end up compiling other things.
        after : list
            Coroutines to wait for before running g++. Typically for writing
            main and header files.

        Returns
        -------
        bool
            Returns True when complete.
        """
        await multiple_waits(after)
        log.info(f"contents of compile dir: {os.listdir(self.tempdir)}")

        cpp_path = os.path.join(self.tempdir, f"{module_name}.cpp")
        env_c_flags = sysconfig.get_config_var('CFLAGS')
        env_c_flags = shlex.split(env_c_flags)
        log.debug(f"CFLAGS: {env_c_flags}")

        cmd = ["g++",
                "-c",
                "-isystem/usr/local/systemc-2.3.3/lib-linux64",
                "-isystem/usr/local/systemc-2.3.3",
                "-isystem/usr/local/systemc-2.3.3/include",
                "-isystem/usr/local/share/verilator/include",
                f"-isystem/usr/local/include/python{sysconfig.get_python_version()}",
                f"-isystem{sysconfig.get_paths()['include']}",
                *env_c_flags,
                *self.CFLAGS,
                cpp_path,
                "-shared",
                "-o", f"{module_name}.o",
                ]

        log.info(f"running cmd {cmd}")
        proc = await asyncio.create_subprocess_exec(*cmd,
                                                    stdout=asyncio.subprocess.PIPE,
                                                    stderr=asyncio.subprocess.PIPE,
                                                    cwd=self.tempdir)

        stdout, stderr = await proc.communicate()
        log.info(stdout.decode("utf8"))
        log.warning(stderr.decode("utf8"))
        if proc.returncode != 0:
            log.error(f"gcc failed with code {proc.returncode}")
            raise RuntimeError(f"gcc failed with code {proc.returncode}")

        return True

    def _get_o(self, module_name, after):
        """After gcc is run, the binary object file is returned.

        Parameters
        ----------
        module_name : str
            The name of the module. Note the .o file type will be
            appended to the string, so is not needed here.
        after : Coroutine
            Coroutine to wait for before the file is ready. Usually
            the process which builds the file.

        Returns
        -------
        BuildArtifact
            The module's binary object.
        """
        return BuildArtifact(f"{module_name}.o", env=self, after=after)

    async def _link(self, main, archive, after, name="main"):
        """Runs linker on main file, with archive.

        Parameters
        ----------
        main : str
            Main object file.
        archive : Union[BuildArtifact, list]
            Archive(s) of object files main is dependent on.
        after : list
            Coroutines to wait for before linking the objects.
        name : Optional[str]
            Name of final runtime, by default "main"

        Returns
        -------
        bool
            Returns True when complete.
        """

        if not isinstance(archive, Iterable):
            archives = [archive]
        else:
            archives = archive

        await multiple_waits(after)
        log.info(f"contents of link dir: {os.listdir(self.tempdir)}")

        main_path = os.path.join(self.tempdir, main)
        archive_paths = [os.path.join(self.tempdir, archive.name)
                         for archive in archives]
        log.info(f"linking: {main_path}, {archive_paths}")

        ld_flags = shlex.split(sysconfig.get_config_var('LDFLAGS'))
        cmd = ["g++",
               "-o",
               name,
               main_path,
               *archive_paths,
               f"-L{find_libpython()}",
               *ld_flags,
               "-L/usr/local/systemc-2.3.3/lib-linux64",
               "-lsystemc",
               "-lpython3.8",
               ]
        log.info(f"running cmd {cmd}")
        proc = await asyncio.create_subprocess_exec(
            *cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=self.tempdir)

        stdout, stderr = await proc.communicate()
        log.info(stdout.decode("utf8"))
        log.warning(stderr.decode("utf8"))

        return True

    def _get_main(self, after, name="main"):
        """Gets main runtime binary.

        Parameters
        ----------
        after : Coroutine
            Coroutine to wait for before getting file.
        name : Optional[str]
            The name of the main runtime file, by default "main"

        Returns
        -------
        BuildArtifact
            The binary main runtime.
        """
        return BuildArtifact(name, self, after=after)

    async def _run_ar(self, objects, after, name="main"):
        """Runs ar to generate .a archive containing all object files.

        Parameters
        ----------
        objects : list
            Objects (.o file names) to put into the archive.
        after : list
            Coroutines to wait for before running the linker.
        name : Optional[str]
            Name of archive, by default "main"

        Returns
        -------
        bool
            Returns True when complete.
        """
        await multiple_waits(after)
        object_paths = [os.path.join(self.tempdir, object_name)
                        for object_name in objects]
        cmd = ["ar",
               "rcsT",
               f"{name}.a",
               *object_paths]
        log.info(f"running {cmd}")
        proc = await asyncio.create_subprocess_exec(*cmd,
                                                    stdout=asyncio.subprocess.PIPE,
                                                    stderr=asyncio.subprocess.PIPE,
                                                    cwd=self.tempdir)

        stdout, stderr = await proc.communicate()
        log.info(stdout.decode("utf8"))
        log.warning(stderr.decode("utf8"))

        return True

    def _get_archive(self, after, name="main"):
        """Get the .a archive.

        Parameters
        ----------
        after : Coroutine
            Coroutine to wait for before the archive is ready.
        name : Optional[str]
            Name of the .a file, by default "main"

        Returns
        -------
        BuildArtifact
            The content of the .a file.
        """
        return BuildArtifact(f"{name}.a", self, after=after)

    async def _write_files(self, files):
        """Save files in the environment's temporary directory.

        Parameters
        ----------
        files : list
            :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>`
            objects containing the files to save.

        Returns
        -------
        bool
            Returns True on completion.

        Raises
        ------
        AttributeError
            Raised if an item in files is not a BuildArtifact.
        """
        for buildartifact in files:
            try:
                with open(os.path.join(self.tempdir, buildartifact.name), "wb") as f:
                    data = await asyncio.wait_for(buildartifact.data, timeout=None)
                    f.write(data)
            except AttributeError:
                log.error(f"{buildartifact} is not a BuildArtifact")
                raise
        return True


    def compile(self, headers, cpp):
        """Compile the C++ file and return the resulting binary.

        Parameters
        ----------
        headers : list
            :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects containing the required header files.
        cpp : BuildArtifact
            File containing the C++ code.

        Returns
        -------
        BuildArtifact
            The compiled binary object file.
        """
        input_files = self._write_files(headers+[cpp])
        top_name = cpp.name.split(".")[0]
        comp_flag = self._run_gcc(module_name=top_name, after=[input_files])
        return self._get_main(after=comp_flag, name=f"{top_name}.o")

    def archive(self, objects, name):
        """Generate archive file containing all binary objects.

        Parameters
        ----------
        objects : list
            :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects containing the binary objects.
        name : str
            Name of the archive file. Note the .a file ending is appended.

        Returns
        -------
        BuildArtifact
            The content of the .a file.
        """
        input_files = self._write_files(objects)
        ar_flag = self._run_ar([o.name for o in objects], after=[
                               input_files], name=name)
        return self._get_main(after=ar_flag, name=f"{name}.a")

    def compile_and_link(self, headers, objects, main_cpp):
        """Run gcc to compile and link and produce a complete runtime binary.

        Parameters
        ----------
        headers : list
            :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects containing the header files.
        objects : list
            :class:`BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects containing the binary objects.
        main_cpp : BuildArtifact
            The C++ file for main.

        Returns
        -------
        BuildArtifact
            The main runtime binary.
        """
        input_files = self._write_files(headers+objects+[main_cpp])

        top_name = main_cpp.name.split(".")[0]
        comp_flag = self._run_gcc(module_name=top_name, after=[input_files])
        link_flag = self._link(
            top_name+".o", archive=objects, after=[comp_flag])
        main = BuildArtifact("main", env=self, after=link_flag)
        return main  # self._get_main(after=link_flag)



import setuptools

install_reqs_path = 'environment/install_requirements.txt'

with open('README.md') as f:
    read_me = f.read()

about_info = {}
with open("deltasimulator/__about__.py") as f:
    exec(f.read(), about_info)


def read_requirements_txt(file_name: str):
    with open(file_name) as f:
        input_lines = f.read().splitlines()

    # Remove trailing/leading whitespace, empty lines and comments
    sanitised = map(str.strip, input_lines)
    sanitised = list(filter(lambda x: not str.startswith(x, '#'), sanitised))
    sanitised = list(filter(None, sanitised))

    return sanitised


setuptools.setup(
    name="deltasimulator",
    version=about_info['__version__'],
    author="Riverlane",
    author_email=about_info['__email__'],
    description=about_info['__short_description__'],
    long_description=read_me,
    long_description_content_type='text/markdown',
    copyright=about_info['__copyright__'],
    license=about_info['__license__'],
    url=about_info['__url__'],
    platforms=about_info['__platforms__'],
    packages=setuptools.find_packages(
        include=['deltasimulator', 'deltasimulator.*']),
    test_suite="test",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: Free for non-commercial use',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: System :: Emulators'
    ],
    install_requires=read_requirements_txt(install_reqs_path),
    python_requires='==3.8.*'
)



import asyncio
from io import StringIO
from os import path
import re
import sysconfig
import textwrap

import dill

from deltalanguage.wiring import PyConstBody, PyInteractiveBody

from .cppenv import CPPEnv
from deltasimulator.build_tools import BuildArtifact, cogify
from deltasimulator.build_tools.utils import multiple_waits

class PythonatorEnv(CPPEnv):
    """Environment for generating SystemC modules of Python nodes.

    :meth:`cogify` is used to automatically generate the Python, C++ and
    header files.

    Parameters
    ----------
    bodies : list
        List of the node bodies, as stored in capnp format.


    .. note::
        Python nodes communicate to each other via SystemC FIFO queues.
        Each queue can store up to 16 items on it. If a node tries to push
        to an already full queue it will be blocked. Similarly, if a node
        tries to read from an empty queue it will be blocked, unless the
        input is optional in which case the node will receive `None`.

    .. note::
        As well as their C++ and header files, a Python script is also
        produced containing the body and data types for the node. This script
        must be stored in the working directory, in order for the nodes to
        load correctly.

    .. note::
        Constant nodes function differently to other Python nodes. Rather
        than evaluating their bodies at runtime, their bodies are evaluated
        as part of the build process and the result hardcoded into the
        output ports. As a result, these nodes do not receive any input and
        have no accompanying Python script.
    """


    def __init__(self, bodies):
        """Initialise the environment

        Parameters
        ----------
        bodies : List[Body]
            List of the node bodies
        """
        super().__init__()
        self._bodies = bodies

    @staticmethod
    def load_port_type(port):
        return dill.loads(port.type)

    @staticmethod
    def as_c_type(df_type):
        """Convert a Deltaflow type to a C type.

        Parameters
        ----------
        df_type : bytes
            The dill serialisation of the Deltaflow type.

        Returns
        -------
        str
            A string describing the SystemC equivalent.
            The format of the string is `sc_dt::sc_bv<{size}>`,
            where `size` is the size of the data type in bits.
        """
        return f"sc_dt::sc_bv<{dill.loads(df_type).size}>"

    @staticmethod
    def get_sysc_port_name(port):
        """Get the name of this port as a SystemC variable.

        Parameters
        ----------
        port
            capnp object describing the port.

        Returns
        -------
        str
            The name of this port in the SystemC module.
            The format used is `sysc_{port_name}`, where
            `port_name` is the name of the port.
        """
        if port.name:
            return f"sysc_{port.name}"
        else:
            return "sysc_return"

    @staticmethod
    def get_module_name(top_p):
        """Get the name of the node's SystemC module.

        Parameters
        ----------
        top_p
            capnp object describing the node.

        Returns
        -------
        str
            The name of the node's module.
            The format used is `{Node_Name}_module`, where
            `Node_Name` is the capitalised version of the name of the node.
        """
        return "_".join([word.capitalize() for word in top_p.name.split("_")])\
            + "_module"


    async def _make_py(self, top_p, df_body, body_type):
        """Makes Python script for the node.

        The Python script contains the node's body as well as data types
        for input and output ports.

        Parameters
        ----------
        top_p
            capnp object describing the node.

        Returns
        -------
        bool
            Returns True on completion.
        """

        # df_ prefix to variables used to avoid namespace clashes with cog
        py_tmpl = """
        import dill
        /*[[[cog
            if body_type is PyInteractiveBody:
                cog.outl(f"import {module_name.lower()}_sysc")
            cog.outl(f"body = dill.loads({df_body})")
            for port in top_p.inPorts:
                cog.outl(f"{self.get_sysc_port_name(port)} = dill.loads({port.type})")
            for port in top_p.outPorts:
                cog.outl(f"{self.get_sysc_port_name(port)} = dill.loads({port.type})")
            if body_type is PyInteractiveBody:
                cog.outl("class SysBridgeNode:")
                cog.outl("    def __init__(self): ")
                inports_list = []
                for port in top_p.inPorts:
                    inports_list.append(port.name)
                inports_str = '[' + '"{0}"'.format('", "'.join(inports_list)) + ']'
                if (len(inports_list) > 0):
                    cog.outl(f"        self.in_queues = dict.fromkeys({inports_str}) ")
                cog.outl("    def send(self,val):")
                if len(top_p.outPorts) > 1:
                    for port in top_p.outPorts:
                        cog.outl(f"        if val.{port.name} is not None:")
                        cog.outl(f"            {module_name.lower()}_sysc.send(\\"{port.name}\\", {self.get_sysc_port_name(port)}.pack(val.{port.name}))")
                else:
                    cog.outl(f"        if val is not None:")
                    cog.outl(f"            {module_name.lower()}_sysc.send(\\"{port.name}\\", {self.get_sysc_port_name(port)}.pack(val))")
                cog.outl("    def receive(self, *args: str):")
                if (len(inports_list) == 0):
                    cog.outl("        raise RuntimeError ")
                cog.outl("        if args: ")
                cog.outl("            in_queue = {name: in_q for name, in_q in self.in_queues.items() if name in args}")
                cog.outl("        else: ")
                cog.outl("            in_queue = self.in_queues ")
                cog.outl("        values = {} ")
                cog.outl("        if in_queue: ")
                cog.outl("             for req in in_queue: ")
                cog.outl(f"                 values[req] = {module_name.lower()}_sysc.receive(req)")
                cog.outl("        if len(values) == 1 and args: ")
                cog.outl("            return values[list(values)[0]] ")
                cog.outl("        else: ")
                cog.outl("            return values")
                cog.outl("node = SysBridgeNode()")
        ]]]*/
        //[[[end]]]
        """
        body_index = top_p.body
        body_file = path.join(self.tempdir, f"{top_p.name}.py")
        module_name = self.get_module_name(top_p)

        with open(body_file, "wb") as py_file:
            py_file.write(cogify(py_tmpl, globals=dict(globals(), **locals())))
        return True

    def _get_py(self, top_p, after):
        """Get the node's Python script.

        Parameters
        ----------
        top_p
            capnp object describing the node.
        after : Coroutine
            Process that needs to complete before the Python script is ready.

        Returns
        -------
        BuildArtifact
            The Python script for this node.
        """
        return BuildArtifact(f"{top_p.name}.py", self, after=after)

    async def _make_h(self, top_p,
                      body_type # pylint: disable=unused-argument
                      ):
        """Generate the Header file for this node.

        Parameters
        ----------
        top_p
            capnp object describing the node.
        body_type
            The type of body this node has. Used to check if the node is a
            constant node or not.

        Returns
        -------
        bool
            Returns True upon completion.
        """

        h_tmpl = """\
        //[[[cog cog.outl(f"#ifndef __{top_p.name.upper()}_MODULE__"); cog.outl(f"#define __{top_p.name.upper()}_MODULE__") ]]]
        //[[[end]]]

        #include <string>
        #include <systemc>
        using namespace sc_core;
        /*[[[cog
            if body_type is not PyConstBody:
                cog.outl('#include "Python.h"')
        ]]]*/
        //[[[end]]]

        //[[[cog cog.outl(f"class {module_name} : public sc_module") ]]]
        //[[[end]]]
        {
        private:
            /*[[[cog
                if body_type is not PyConstBody:
                    cog.outl("PyObject *pBody, *pName, *pModule, *pyC, *pExit;")
                if body_type is PyInteractiveBody:
                    cog.outl("PyObject *runtimeModule, *pNode;")
                    cog.outl("static PyObject* sc_receive(PyObject *self, PyObject *args);")
                    cog.outl("static PyObject* sc_send(PyObject *self, PyObject *args);")
                    cog.outl("static PyObject* PyInit_sysc(void);")
                    cog.outl("static PyMethodDef SysCMethods[];")
                    cog.outl("static PyModuleDef SysCModule;")
                    cog.outl(f"static {module_name}* singleton;")
                    cog.outl("void init();")
                    cog.outl("void store();")
                    cog.outl("void run();")
            ]]]*/
            //[[[end]]]
            uint64_t no_ins, no_outs;
            /*[[[cog
                if body_type is not PyConstBody:
                    for port in top_p.inPorts:
                        cog.outl(f"PyObject* type_{self.get_sysc_port_name(port)};")
                        cog.outl(f"PyObject* get_{self.get_sysc_port_name(port)}();")
                        cog.outl(f"{self.as_c_type(port.type)} bits_{self.get_sysc_port_name(port)};")
                for port in top_p.outPorts:
                    if body_type is not PyConstBody:
                        cog.outl(f"PyObject* type_{self.get_sysc_port_name(port)};")
                        cog.outl(f"{self.as_c_type(port.type)} bits_{self.get_sysc_port_name(port)};")
                    cog.outl(f"void set_{self.get_sysc_port_name(port)}();")
            ]]]*/
            //[[[end]]]
        public:
            uint64_t no_inputs, no_outputs;
            /*[[[cog
                    if body_type is not PyConstBody:
                        for port in top_p.inPorts:
                            cog.outl(f"sc_fifo<{self.as_c_type(port.type)}>* {self.get_sysc_port_name(port)};")
                    for port in top_p.outPorts:
                        cog.outl(f"sc_fifo<{self.as_c_type(port.type)}>* {self.get_sysc_port_name(port)};")
            ]]]*/
            //[[[end]]]
            int get_no_inputs() const;
            int get_no_outputs() const;
            void body();
            //[[[cog cog.outl(f"SC_HAS_PROCESS({module_name});"); cog.outl(f"{module_name}(sc_module_name name);") ]]]
            //[[[end]]]
        };
        #endif
        """

        module_name = self.get_module_name(top_p)
        h_name = path.join(self.tempdir, f"{top_p.name}.h")

        with open(h_name, "wb") as out_file:
            out_file.write(cogify(h_tmpl, globals=dict(globals(), **locals())))
        return True

    def _get_h(self, top_p, after):
        """Get the Header file for the node.

        Parameters
        ----------
        top_p
            capnp object describing the node.
        after : Coroutine
            The process that constructs the Header file.

        Returns
        -------
        bool
            Returns True upon completion.
        """
        return BuildArtifact(f"{top_p.name}.h", self, after=after)

    async def _make_cpp(self, top_p,
                        body_type # pylint: disable=unused-argument
                        ):
        """Construct the node's SystemC module.

        Parameters
        ----------
        top_p
            capnp object describing the node.
        body_type
            The type of body this node has. Used to check if the node is a
            constant node or not.

        Returns
        -------
        bool
            Returns True upon completion.
        """
        cpp_tmpl = """\
        //[[[cog cog.outl('#include "' + top_p.name + '.h"') ]]]
        //[[[end]]]

        /*[[[cog
        if body_type is PyInteractiveBody:
            cog.outl(f"{module_name}* {module_name}::singleton = nullptr;")
        ]]]*/
        //[[[end]]]
        //[[[cog cog.outl(f"{module_name}::{module_name}(sc_module_name name): sc_module(name) {{") ]]]
        //[[[end]]]
            /*[[[cog
            if body_type is PyInteractiveBody:
                cog.outl("if (singleton == nullptr) {")
                cog.outl("singleton = this;")
                cog.outl("} else {")
                cog.outl(f"std::cerr << \\"attempted to construct multiple python for {module_name} modules - not supported!\\" << std::endl;")
                cog.outl("exit(-1);")
                cog.outl("}")
            if body_type is PyConstBody:
                cog.outl('no_ins = 0;')
            else:
                cog.outl(f'no_ins = {len(top_p.inPorts)};')
            cog.outl(f'no_outs = {len(top_p.outPorts)};')
            if body_type is PyInteractiveBody:
                cog.outl(f'const char* sc_module_name = "{module_name.lower()}_sysc";')
                cog.outl('PyImport_AddModule(sc_module_name);')
                cog.outl('PyObject* sys_modules = PyImport_GetModuleDict();')
                cog.outl('PyObject* module = PyInit_sysc();')
                cog.outl('PyDict_SetItemString(sys_modules, sc_module_name, module);')
            if body_type is not PyConstBody:
                cog.outl(f'this->pName = PyUnicode_DecodeFSDefault("{top_p.name}");')
                cog.outl("this->pModule = PyImport_Import(this->pName);")
                cog.outl("if (this->pModule == NULL) {")
                cog.outl("if (PyErr_Occurred()) PyErr_Print();")
                cog.outl(f'std::cout << "failed to import {top_p.name} python module." << std::endl;')
                cog.outl("exit(-1);")
                cog.outl("}")
                cog.outl("this->pBody = PyObject_GetAttrString(this->pModule, \\"body\\");")
                cog.outl("if (this->pBody == NULL) {")
                cog.outl("if (PyErr_Occurred()) PyErr_Print();")
                cog.outl(f'std::cout << "failed to import {top_p.name} python body." << std::endl;')
                cog.outl("exit(-1);")
                cog.outl("}")
                if body_type is PyInteractiveBody:
                    cog.outl("this->pNode = PyObject_GetAttrString(this->pModule, \\"node\\");")
                    cog.outl("if (this->pBody == NULL) {")
                    cog.outl("if (PyErr_Occurred()) PyErr_Print();")
                    cog.outl(f'std::cout << "failed to import {top_p.name} python interactive node." << std::endl;')
                    cog.outl("exit(-1);")
                    cog.outl("}")
                cog.outl("PyObject *runtimeModule = PyImport_Import(PyUnicode_DecodeFSDefault(\\"deltalanguage.runtime\\"));")
                cog.outl("if (runtimeModule == NULL) {")
                cog.outl("if (PyErr_Occurred()) PyErr_Print();")
                cog.outl(f'std::cout << "failed to import deltalanguage runtime in {top_p.name}." << std::endl;')
                cog.outl("exit(-1);")
                cog.outl("}")
                cog.outl("this->pExit = PyObject_GetAttrString(runtimeModule, \\"DeltaRuntimeExit\\");")
                cog.outl("if (this->pExit == NULL) {")
                cog.outl("if (PyErr_Occurred()) PyErr_Print();")
                cog.outl(f'std::cout << "failed to import exit exception object from deltalanguage runtime in {top_p.name}." << std::endl;')
                cog.outl("exit(-1);")
                cog.outl("}")
                for port in top_p.inPorts:
                    cog.outl(f"this->type_{self.get_sysc_port_name(port)} = PyObject_GetAttrString(this->pModule, \\"{self.get_sysc_port_name(port)}\\");")
                    cog.outl(f"if (this->type_{self.get_sysc_port_name(port)} == NULL){{")
                    cog.outl("if (PyErr_Occurred()) PyErr_Print();")
                    cog.outl(f"std::cout << \\"failed to import type for in port {port.name} in {top_p.name}.\\" << std::endl;")
                    cog.outl("exit(-1);")
                    cog.outl("}")
                    cog.outl(f"{self.get_sysc_port_name(port)} = NULL;")
                for port in top_p.outPorts:
                    if body_type is not PyConstBody:
                        cog.outl(f"this->type_{self.get_sysc_port_name(port)} = PyObject_GetAttrString(this->pModule, \\"{self.get_sysc_port_name(port)}\\");")
                        cog.outl(f"if (this->type_{self.get_sysc_port_name(port)} == NULL){{")
                        cog.outl("if (PyErr_Occurred()) PyErr_Print();")
                        if port.name:
                            cog.outl(f"std::cout << \\"failed to import type for out port {port.name} in {top_p.name}.\\" << std::endl;")
                        else:
                            cog.outl(f"std::cout << \\"failed to import return type in {top_p.name}.\\" << std::endl;")
                        cog.outl("exit(-1);")
                        cog.outl("}")
                    cog.outl(f"{self.get_sysc_port_name(port)} = NULL;")
            if body_type is not PyConstBody:
                cog.outl("Py_XDECREF(this->pName);")
                cog.outl("Py_XDECREF(this->pModule);")
            ]]]*/
            //[[[end]]]
            SC_THREAD(body);
        }

        /*[[[cog
            if body_type not in [PyConstBody, PyInteractiveBody]:
                for port in top_p.inPorts:
                    port_type = self.load_port_type(port)
                    cog.outl(f"PyObject* {module_name}::get_{self.get_sysc_port_name(port)}(){{")
                    cog.outl(f"    if ({self.get_sysc_port_name(port)} == NULL) return Py_None;")
                    if port.optional:
                        cog.outl(f"    if (!{self.get_sysc_port_name(port)}->nb_read(bits_{self.get_sysc_port_name(port)})) return Py_None;")
                    else:
                        cog.outl(f"    bits_{self.get_sysc_port_name(port)} = {self.get_sysc_port_name(port)}->read();")
                    cog.outl(f"    return PyObject_CallMethodObjArgs(this->type_{self.get_sysc_port_name(port)}, PyUnicode_FromString(\\"unpack\\"), PyBytes_FromStringAndSize(bits_{self.get_sysc_port_name(port)}.to_string().c_str(), {port_type.size}), NULL);")
                    cog.outl("};")
            if body_type is PyInteractiveBody:
                cog.outl(f"PyMethodDef {module_name}::SysCMethods[] = {{")
                cog.outl("{\\"receive\\", sc_receive, METH_VARARGS, \\"Receives data from the systemC interface\\"},")
                cog.outl("{\\"send\\", sc_send, METH_VARARGS, \\"Sends data to the systemC interface\\"},")
                cog.outl("{NULL, NULL, 0, NULL}")
                cog.outl("};")

                cog.outl(f"PyModuleDef {module_name}::SysCModule = {{")
                cog.outl(f"PyModuleDef_HEAD_INIT, \\"{module_name.lower()}_sysc\\", NULL, -1, SysCMethods,")
                cog.outl("NULL, NULL, NULL, NULL")
                cog.outl("};")

                cog.outl(f"PyObject* {module_name}::PyInit_sysc(void)")
                cog.outl("{")
                cog.outl(f"std::cout << \\"{module_name}::PyInit_sysc() called \\" << std::endl;")
                cog.outl("return PyModule_Create(&SysCModule);")
                cog.outl("}")

                cog.outl(f"PyObject* {module_name}::sc_receive(PyObject *self, PyObject *args){{")
                cog.outl('    const char * inport; ')
                cog.outl('    bool retval; ')
                for port in top_p.inPorts:
                    cog.outl(f"{self.as_c_type(port.type)} bv_{self.get_sysc_port_name(port)};")
                cog.outl('    if (!PyArg_ParseTuple(args, "s", &inport )) {')
                cog.outl('        std::cout << "sc_receive::ERROR" << std::endl;')
                cog.outl('        return NULL;')
                cog.outl('    }')
                cog.outl('    std::string ins (inport);')
                num_ports = len(top_p.inPorts)
                for port in top_p.inPorts:
                    port_type = self.load_port_type(port)
                    if num_ports > 1:
                        cog.outl(f"    if (ins == \\"{port.name}\\") {{")
                    if port.optional:
                        cog.outl(f"      retval = singleton->{self.get_sysc_port_name(port)}->nb_read(bv_{self.get_sysc_port_name(port)});")
                        cog.outl("      if (retval == false){")
                        cog.outl("          return Py_None;")
                        cog.outl("      }")
                    else :
                        cog.outl(f"      singleton->{self.get_sysc_port_name(port)}->read(bv_{self.get_sysc_port_name(port)});")
                    cog.outl(f"return PyObject_CallMethodObjArgs(singleton->type_{self.get_sysc_port_name(port)}, PyUnicode_FromString(\\"unpack\\"), PyBytes_FromStringAndSize(bv_{self.get_sysc_port_name(port)}.to_string().c_str(), {port_type.size}), NULL);")
                    if num_ports > 1:
                        cog.outl("}")
                cog.outl('    PyErr_SetString(PyExc_TypeError, "Unrecognized argument");')
                cog.outl('    return (PyObject *) NULL;')
                cog.outl("}")

                cog.outl(f"PyObject* {module_name}::sc_send(PyObject *self, PyObject *args){{")
                cog.outl('    const char * outport;')
                cog.outl('    const char * data;')
                cog.outl('    if (!PyArg_ParseTuple(args, "ss*", &outport, &data)) {')
                cog.outl('        std::cout << "sc_send::ERROR" << std::endl;')
                cog.outl('        return NULL;')
                cog.outl('    }')
                cog.outl('    singleton->wait(1, SC_NS);')
                cog.outl('    std::string outs (outport);')

                num_ports = len(top_p.outPorts)
                for port in top_p.outPorts:
                    if num_ports > 1:
                        cog.outl(f"    if (outs == \\"{port.name}\\"){{")
                        cog.outl(f"            singleton->{self.get_sysc_port_name(port)}->write(data);")
                        cog.outl("    }")
                    else:
                        cog.outl(f"    singleton->{self.get_sysc_port_name(port)}->write(data);")
                cog.outl('    return Py_None;')
                cog.outl("}")
        ]]]*/
        //[[[end]]]

        /*[[[cog
            if body_type is not PyInteractiveBody:
                if body_type is PyConstBody:
                    body = dill.loads(self._bodies[top_p.body].python.dillImpl)
                    val = body.eval()
                for port in top_p.outPorts:
                    port_type = self.load_port_type(port)
                    cog.outl(f"void {module_name}::set_{self.get_sysc_port_name(port)}(){{")
                    cog.outl(f"    if ({self.get_sysc_port_name(port)} != NULL){{")
                    if body_type is PyConstBody:
                        if port.name:
                            port_val = val[port.name]
                        else:
                            port_val = val
                        if port_val is not None:
                            cog.outl(f'        {self.get_sysc_port_name(port)}->nb_write("{port_type.pack(port_val).decode("ascii")}");')
                        else:
                            raise ValueError(f'None returned by constant node {top_p.name}.')
                    else:
                        if port.name:
                            cog.outl(f"        PyObject* pyRet = PyObject_GetAttrString(this->pyC,\\"{port.name}\\");")
                        else:
                            cog.outl(f"        PyObject* pyRet = this->pyC;")
                        cog.outl("        if (pyRet != Py_None){")
                        cog.outl(f"            PyObject* pyBits = PyObject_CallMethodObjArgs(this->type_{self.get_sysc_port_name(port)}, PyUnicode_FromString(\\"pack\\"), pyRet, NULL);")
                        cog.outl("             PyObject* pyErr = PyErr_Occurred();")
                        cog.outl("             if (pyErr != NULL) {")
                        cog.outl("                 PyErr_Print();")
                        cog.outl("                 PyErr_Clear();")
                        cog.outl("                 exit(-1);")
                        cog.outl("             }")
                        cog.outl(f"            char* bitsRet = PyBytes_AsString(pyBits);")
                        cog.outl(f"            {self.get_sysc_port_name(port)}->write(bitsRet);")
                        cog.outl("        }")
                    cog.outl("    }")
                    cog.outl("};")
        ]]]*/
        //[[[end]]]

        //[[[cog cog.outl(f'void {module_name}::body(){{') ]]]
        //[[[end]]]
            /*[[[cog
            if body_type is not PyInteractiveBody:
                cog.outl('while (true) {')
            ]]]*/
            //[[[end]]]
                /*[[[cog

                    if body_type is not PyConstBody:
                        if body_type is PyInteractiveBody:
                            cog.outl('this->pyC = PyObject_CallMethodObjArgs(this->pBody, PyUnicode_FromString("eval"), this->pNode , NULL);')
                        else:
                            if top_p.inPorts:
                                args = [f'get_{self.get_sysc_port_name(port)}()' for port in top_p.inPorts]
                                cog.outl('this->pyC = PyObject_CallMethodObjArgs(this->pBody, PyUnicode_FromString("eval"),' + ",".join(args) +' ,NULL);')
                            else:
                                cog.outl('this->pyC = PyObject_CallMethod(this->pBody, "eval", NULL);')
                        cog.outl("PyObject* pyErr = PyErr_Occurred();")
                        cog.outl("if (pyErr != NULL) {")
                        cog.outl("    if (PyErr_ExceptionMatches(this->pExit)) {")
                        cog.outl("        PyErr_Clear();")
                        cog.outl("        sc_stop();")
                        if body_type is not PyInteractiveBody:
                            cog.outl("        break;")
                        cog.outl("    } else {")
                        cog.outl("        PyErr_Print();")
                        cog.outl("        PyErr_Clear();")
                        cog.outl("        exit(-1);")
                        cog.outl("    }")
                        cog.outl("}")
                ]]]*/
                //[[[end]]]
                /*[[[cog
                    if body_type is not PyInteractiveBody:
                        if body_type is PyConstBody:
                            for port in top_p.outPorts:
                                cog.outl(f'set_{self.get_sysc_port_name(port)}();')
                        else:
                            cog.outl("if (this->pyC != Py_None) {")
                            for port in top_p.outPorts:
                                cog.outl(f'    set_{self.get_sysc_port_name(port)}();')
                            cog.outl("}")
                ]]]*/
                //[[[end]]]
            /*[[[cog
            if body_type is not PyInteractiveBody:
                cog.outl('    wait(1, SC_NS);')
                cog.outl('}')
            ]]]*/
            //[[[end]]]
        };

        //[[[cog cog.outl(f"int {module_name}::get_no_inputs() const") ]]]
        //[[[end]]]
        {
            return no_ins;
        };

        //[[[cog cog.outl(f"int {module_name}::get_no_outputs() const") ]]]
        //[[[end]]]
        {
            return no_outs;
        };
        """

        module_name = self.get_module_name(top_p)
        cpp_name = path.join(self.tempdir, f"{top_p.name}.cpp")
        with open(cpp_name, "wb") as out_file:
            out_file.write(cogify(cpp_tmpl, globals=dict(globals(), **locals())))
        return True

    def _get_cpp(self, top_p, after):
        """Get the C++ code containing the node's SystemC module after
        it has been built.

        Parameters
        ----------
        top_p
            capnp object describing the node.
        after : Coroutine
            Process which builds the C++ code.

        Returns
        -------
        BuildArtifact
            The C++ code implementing the node's SystemC module.
        """
        return BuildArtifact(f"{top_p.name}.cpp", self, after=after)

    async def _build_objects(self, top_p, after):
        """Build the object file using gcc.

        Parameters
        ----------
        top_p
            capnp object describing the node. Used to get the node's name.
        after : list
            :class:`Coroutine` objects that need to finish before the
            binary object can be built. This should include the processes
            for building the node's header and C++ files.

        Returns
        -------
        Bool
            Returns True once build is complete.
        """
        done = await self._run_gcc(top_p.name, after)
        return done

    def _get_binary(self, top_p, after):
        """Get the binary object for the node.

        Parameters
        ----------
        top_p
            capnp object describing the node. Used to get the node's name.
        after : Coroutine
            Build process that needs to complete before the binary object
            is ready.

        Returns
        -------
        BuildArtifact
            The built binary object for the node.
        """
        return self._get_o(top_p.name, after)

    def pythonate(self, top_p):
        """Generates all the build outputs for this node.

        Parameters
        ----------
        top_p
            capnp object describing the node.

        Returns
        -------
        dict
            A map of strings to `BuildArtifact<deltasimulator.build_tools.BuildArtifact>` objects containing different
            parts of the node's SystemC implementation:

            - "cpp": the C++ file
            - "h": the header file
            - "py": the Python file (only exists if the node is not constant)
            - "o": the built binary object
        """
        body_class = self._bodies[top_p.body].which()
        if "python" in body_class:
            body = self._bodies[top_p.body].python.dillImpl
        if "interactive" in body_class:
            body = self._bodies[top_p.body].interactive.dillImpl
        body_types = {b"PyConstBody": PyConstBody,
                      b"PyInteractiveBody": PyInteractiveBody}
        match = re.search(b"Py(Const|Interactive)Body", body)
        if match:
            body_type = body_types[match.group(0)]
        else:
            body_type = None
        make_cpp = self._make_cpp(top_p, body_type)
        cpp = self._get_cpp(top_p, after=make_cpp)
        make_h = self._make_h(top_p, body_type)
        h = self._get_h(top_p, after=make_h)
        if body_type is not PyConstBody:
            make_py = self._make_py(top_p, body, body_type)
            py = self._get_py(top_p, after=make_py)
        build_objects = self._build_objects(top_p, after=[make_cpp, make_h])
        binary = self._get_binary(top_p, after=build_objects)
        # return a set of build artifacts
        built_artifacts = dict()
        built_artifacts["cpp"] = cpp
        built_artifacts["h"] = h
        if body_type is not PyConstBody:
            built_artifacts["py"] = py
        built_artifacts["o"] = binary
        return built_artifacts



import cogapp
from io import StringIO
import textwrap


def cogify(template, globals=None, cog=cogapp.cogapp.Cog(), verbose=False):
    """
    Passes the string str through the cog template engine.

    Parameters
    ----------
    template : str
        cog template string.
    globals : Optional[dict]
        dict for variable lookup within the cog template.
        pass globals=locals() at the call site to make local names
        available to the template. Otherwise no names will be present.
    cog
        Cog instance to use: pass in a alternative if you wish to change
        the behaviour of the engine.

    Returns
    -------
    bytes
        a bytes object of the string encoded as utf-8.


    .. note::
        Leading whitespace is stripped from every line: you can allign a
        multiline string to pass to this function with the containing block and
        it will be fixed up.

    .. note::
        cog.outl() is required for getting strings into the output area.

    .. warning::
        Ensure there are no explicit newlines within a cog python block.
        If strings with newlines are present, they break python code formatting.
    """
    if globals is None:
        globals = dict()
    template = textwrap.dedent(template)
    fOld = StringIO(template)
    fNew = StringIO()
    try:
        if not verbose:
            cog.options.bDeleteCode = True
        cog.processFile(fOld, fNew, fname=None, globals=globals)
        return fNew.getvalue().encode("utf-8")
    except cogapp.cogapp.CogUserException as ex:
        if "DeltaRuntimeExit" in ex.args[0]:
            raise ValueError(
                "DeltaRuntimeExit raised during build process.") from ex
        else:
            raise ValueError("Exception raised during build process.") from ex



from .__about__ import (
    __license__,
    __copyright__,
    __url__,
    __contributors__,
    __version__,
    __doc__
)

__all__ = [
    "__license__",
    "__copyright__",
    "__url__",
    "__contributors__",
    "__version__",
    "__doc__"
]
